import asyncio
import logging
from decimal import Decimal
from typing import List, Optional

import a_sync
from brownie import ZERO_ADDRESS
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20
from y.contracts import has_method
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import call_reverted
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory')
async def is_pie(token: AnyAddressType) -> bool:
    return await has_method(token, "getCap()(uint)", sync=False)

@a_sync.a_sync(default='sync')
async def get_price(pie: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
    tvl, total_supply = await asyncio.gather(
        get_tvl(pie, block, skip_cache=skip_cache),
        ERC20(pie, asynchronous=True).total_supply_readable(block),
    )
    return UsdPrice(tvl / total_supply)

async def get_tokens(pie_address: Address, block: Optional[Block] = None) -> List[ERC20]:
    return [ERC20(t) for t in await Call(pie_address, ['getTokens()(address[])'], block_id=block)]

async def get_bpool(pie_address: Address, block: Optional[Block] = None) -> Address:
    try:
        bpool = await raw_call(pie_address, 'getBPool()', output='address', block=block, sync=False)
        return bpool if bpool != ZERO_ADDRESS else pie_address
    except Exception as e:
        if not call_reverted(e):
            raise
        return pie_address

async def get_tvl(pie_address: Address, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdValue:
    tokens: List[ERC20]
    pool, tokens = await asyncio.gather(get_bpool(pie_address, block), get_tokens(pie_address, block))
    return await a_sync.map(get_value, tokens, bpool=pool, block=block, skip_cache=skip_cache).sum(pop=True, sync=False)

async def get_balance(bpool: Address, token: ERC20, block: Optional[Block] = None) -> Decimal:
    balance, scale = await asyncio.gather(Call(token.address, ['balanceOf(address)(uint)', bpool], block_id=block), token.__scale__)
    return Decimal(balance) / scale

async def get_value(bpool: Address, token: ERC20, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdValue:
    balance, price = await asyncio.gather(get_balance(bpool, token, block), token.price(block, skip_cache=skip_cache, sync=False))
    return UsdValue(balance * Decimal(price))




import asyncio
import logging
from collections import defaultdict
from contextlib import suppress
from enum import IntEnum
from typing import (Any, AsyncIterator, Awaitable, Callable, Dict, List, NewType, 
                    Optional, Tuple, TypeVar, Union)

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import ZERO_ADDRESS, chain
from brownie.convert.datatypes import EthAddress
from brownie.network.contract import ContractCall, ContractTx, OverloadedMethod
from brownie.network.event import _EventItem
from hexbytes import HexBytes
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import constants, contracts
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.contracts import Contract
from y.datatypes import Address, AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import ContractNotVerified, TokenNotFound
from y.networks import Network
from y.prices.dex.balancer._abc import BalancerABC, BalancerPool
from y.utils.cache import a_sync_ttl_cache
from y.utils.events import ProcessedEvents
from y.utils.logging import get_price_logger


BALANCER_V2_VAULTS = {
    Network.Mainnet: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Fantom: [
        '0x20dd72Ed959b6147912C2e529F0a0C651c33c9ce',
    ],
    Network.Polygon: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Arbitrum: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
    Network.Base: [
        '0xBA12222222228d8Ba445958a75a0704d566BF2C8',
    ],
}.get(chain.id, [])

MESSED_UP_POOLS = {
    Network.Mainnet: [
        # NOTE: this was the first ever balancer "pool" and isn't actually a pool
        "0xF3799CBAe9c028c1A590F4463dFF039b8F4986BE",
    ],
}.get(chain.id, [])

T = TypeVar("T")

PoolId = NewType('PoolId', bytes)
PoolBalances = Dict[ERC20, WeiBalance]

logger = logging.getLogger(__name__)

class PoolSpecialization(IntEnum):
    ComposableStablePool = 0
    WeightedPool = 1
    WeightedPool2Tokens = 2
    # This is a weird one
    CronV1Pool = -1

    @staticmethod
    def with_immutable_tokens() -> List["PoolSpecialization"]:
        return [
            PoolSpecialization.ComposableStablePool,
            PoolSpecialization.WeightedPool,
            PoolSpecialization.WeightedPool2Tokens,
            PoolSpecialization.CronV1Pool,
        ]


class BalancerV2Vault(ContractBase):
    def __init__(self, address: AnyAddressType, asynchronous: bool = False) -> None:
        super().__init__(address, asynchronous=asynchronous)
        self._events = BalancerEvents(self, addresses=address, topics=['0x3c13bc30b8e878c53fd2a36b679409c073afd75950be43d8858768e956fbc20e'])
        if not self._is_cached:
            # we need the contract cached so we can decode logs correctly
            self.contract
    
    @stuck_coro_debugger
    async def pools(self, block: Optional[Block] = None) -> AsyncIterator["BalancerV2Pool"]:
        async for pool in self._events.events(to_block=block):
            yield pool

    @stuck_coro_debugger
    async def pools_for_token(self, token: Address, block: Optional[Block] = None) -> AsyncIterator["BalancerV2Pool"]:
        async for pool, tokens in BalancerV2Pool.tokens.map(block=block).map(self.pools(block=block), pop=True):
            if token in tokens:
                logger.debug("%s contains %s", pool, token)
                yield pool
                
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_tokens(self, pool_id: HexBytes, block: Optional[Block] = None):
        contract = await contracts.Contract.coroutine(self.address)
        return await contract.getPoolTokens.coroutine(pool_id, block_identifier = block)
    
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_pool_info(self, poolids: Tuple[HexBytes,...], block: Optional[Block] = None) -> List[Tuple]:
        contract = await contracts.Contract.coroutine(self.address)
        return await contract.getPoolTokens.map(poolids, block_identifier=block)
    
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> "BalancerV2Pool":
        balance_tasks: a_sync.TaskMapping[BalancerV2Pool, Optional[WeiBalance]]

        logger = get_price_logger(token_address, block, extra='balancer.v2')

        balance_tasks = BalancerV2Pool.get_balance.map(token_address=token_address, block=block)
        balances_aiterator = balance_tasks.map(self.pools_for_token(token_address, block=block), pop=True)        
        async for pool, balance in balances_aiterator.filter(_lookup_balance_from_tuple).sort(key=_lookup_balance_from_tuple, reverse=True):
            logger.debug("deepest pool %s balance %s", pool, balance)
            return pool

class BalancerEvents(ProcessedEvents[Tuple[HexBytes, EthAddress, Block]]):
    threads = a_sync.PruningThreadPoolExecutor(6)
    __slots__ = "asynchronous", 
    def __init__(self, vault: BalancerV2Vault, *args, asynchronous: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.vault = vault
        self.asynchronous = asynchronous
        self.__tasks = []
    def _include_event(self, event: _EventItem) -> Awaitable[bool]:
        if event['poolAddress'] in MESSED_UP_POOLS:
            return False
        # NOTE: For some reason the Balancer fork on Fantom lists "0x3e522051A9B1958Aa1e828AC24Afba4a551DF37d"
        #       as a pool, but it is not a contract. This handler will prevent it and future cases from causing problems.
        # NOTE: this isn't really optimized as it still runs semi-synchronously but its better than what was had previously
        return self.threads.run(contracts.is_contract, event['poolAddress'])
    def _process_event(self, event: _EventItem) -> "BalancerV2Pool":
        try:
            specialization = PoolSpecialization(event['specialization'])
        except ValueError:
            specialization = None
        pool = BalancerV2Pool(
            address = event['poolAddress'], 
            id = HexBytes(event['poolId']), 
            specialization=specialization,
            vault=self.vault,
            _deploy_block = event.block_number,
            asynchronous = self.asynchronous, 
        )
        # lets get this cached into memory now
        task = asyncio.create_task(pool.tokens(sync=False))
        self.__tasks.append(task)
        task.add_done_callback(self._task_done_callback)
        return pool
    def _get_block_for_obj(self, pool: "BalancerV2Pool") -> int:
        return pool._deploy_block
    def _task_done_callback(self, t: asyncio.Task):
        self.__tasks.remove(t)
        if not t.cancelled():
            # get the exc so it doesn't log, it will come up later
            t.exception()


class BalancerV2Pool(BalancerPool):
    """A pool from Balancer Protocol v2"""
    # internal variables to save calls in some instances
    # they do not necessarily reflect real life at all times
    # defaults are stored as class vars to keep instance dicts smaller
    __nonweighted: bool = False
    __tokens: Tuple[ERC20, ...] = None
    __weights: List[int] = None
    def __init__(
        self, 
        address: AnyAddressType,
        *,
        id: Optional[HexBytes] = None,
        specialization: Optional[PoolSpecialization] = None,
        vault: Optional[BalancerV2Vault] = None,
        asynchronous: bool = False, 
        _deploy_block: Optional[Block] = None,
    ):
        super().__init__(address, asynchronous=asynchronous, _deploy_block=_deploy_block)
        if id is not None:
            self.id = id
        if specialization is not None:
            self.pool_type = specialization
        if vault is not None:
            self.vault = vault

    @a_sync.aka.cached_property
    async def id(self) -> PoolId:
        return await Call(self.address, ['getPoolId()(bytes32)'])
    __id__: HiddenMethodDescriptor[Self, PoolId]
    
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def vault(self) -> Optional[BalancerV2Vault]:
        with suppress(ContractLogicError):
            vault = await Call(self.address, ['getVault()(address)'])
            if vault == ZERO_ADDRESS:
                return None
            elif vault:
                return BalancerV2Vault(vault, asynchronous=True)
        # in earlier web3 versions, we would get `None`. More recently, we get ContractLogicError. This handles both
        if chain.id == Network.Mainnet and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi? 
            return BalancerV2Vault("0xBA12222222228d8Ba445958a75a0704d566BF2C8", asynchronous=True)
    __vault__: HiddenMethodDescriptor[Self, Optional[BalancerV2Vault]]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pool_type(self) -> Union[PoolSpecialization, int]:
        vault = await self.__vault__
        if vault is None:
            raise ValueError(f"{self} has no vault") from None
        elif poolid := await self.__id__:
            _, specialization = await vault.contract.getPool.coroutine(poolid)
        elif chain.id == Network.Mainnet and await self.__build_name__ == "CronV1Pool":
            # NOTE: these `CronV1Pool` tokens ARE balancer pools but don't match the expected pool abi? 
            return PoolSpecialization.CronV1Pool
        else:
            raise ValueError(f"{self} has no poolid") from None
        try:
            return PoolSpecialization(specialization)
        except ValueError:
            if self.address not in _warned:
                with suppress(ContractNotVerified):
                    logger.warning(
                        "ypricemagic does not recognize this pool type, please add `%s = %s` to %s.PoolSpecialization (pool=%s)", 
                        await self.__build_name__, specialization, __name__, self.address
                    )
                _warned.add(self.address)
            return specialization
    __pool_type__: HiddenMethodDescriptor[Self, Optional[PoolSpecialization]]
        
    @stuck_coro_debugger
    async def get_tvl(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdValue]:
        if balances := await self.get_balances(block=block, skip_cache=skip_cache, sync=False):
            # overwrite ref to big obj with ref to little obj
            balances = iter(tuple(balances.values()))
            return UsdValue(await WeiBalance.value_usd.sum(balances, sync=False))

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def get_balances(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Dict[ERC20, WeiBalance]:
        vault = await self.__vault__
        if vault is None:
            return {}
        tokens, balances, lastChangedBlock = await vault.get_pool_tokens(await self.__id__, block=block, sync=False)
        return {
            ERC20(token, asynchronous=self.asynchronous): WeiBalance(balance, token, block=block, skip_cache=skip_cache)
            for token, balance in zip(tokens, balances)
            # NOTE: some pools include themselves in their own token list, and we should ignore those
            if token != self.address
        }
    
    async def get_balance(self, token_address: Address, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[WeiBalance]:
        if info := await self.get_balances(block=block, sync=False):
            try:
                return info[token_address]
            except KeyError:
                raise TokenNotFound(token_address, self) from None

    @stuck_coro_debugger
    async def get_token_price(self, token_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        get_balances_coro = self.get_balances(block=block, skip_cache=skip_cache, sync=False)
        if self.__nonweighted:
            # this await will return immediately once cached
            token_balances = await get_balances_coro
            weights = self.__weights
        else:
            token_balances, weights = await asyncio.gather(get_balances_coro, self.weights(block=block, sync=False))
        pool_token_info = list(zip(token_balances.keys(), token_balances.values(), weights))
        for pool_token, token_balance, token_weight in pool_token_info:
            if pool_token == token_address:
                break

        paired_token_balance: Optional[WeiBalance] = None
        for pool_token, balance, weight in pool_token_info:
            if pool_token in constants.STABLECOINS:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif pool_token == constants.WRAPPED_GAS_COIN:
                paired_token_balance, paired_token_weight = balance, weight
                break
            elif len(pool_token_info) == 2 and pool_token != token_address:
                paired_token_balance, paired_token_weight = balance, weight
                break
        
        if paired_token_balance is None:
            return None

        token_value_in_pool, token_balance_readable = await asyncio.gather(paired_token_balance.__value_usd__, token_balance.__readable__)
        token_value_in_pool /= paired_token_weight * token_weight
        return UsdPrice(token_value_in_pool / token_balance_readable) 

    # NOTE: We can't cache this as a cached property because some balancer pool tokens can change. Womp
    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def tokens(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Tuple[ERC20, ...]:
        if self.__tokens:
            return self.__tokens
        tokens = tuple((await self.get_balances(block=block, skip_cache=skip_cache, sync=False)).keys())
        if await self.__pool_type__ in PoolSpecialization.with_immutable_tokens():
            self.__tokens = tokens
        return tokens

    @a_sync_ttl_cache
    @stuck_coro_debugger
    async def weights(self, block: Optional[Block] = None) -> List[int]:
        contract = await Contract.coroutine(self.address)
        try:
            return await contract.getNormalizedWeights.coroutine(block_identifier = block)
        except AttributeError:
            # Contract has no method `getNormalizedWeights`
            self.__nonweighted = True
            num_tokens = len(await self.tokens(block=block, sync=False))
            self.__weights = [10 ** 18 // num_tokens] * num_tokens
            return self.__weights


class BalancerV2(BalancerABC[BalancerV2Pool]):

    _pool_type = BalancerV2Pool

    _check_methods = ('getPoolId()(bytes32)','getPausedState()((bool,uint,uint))','getSwapFeePercentage()(uint)')

    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        self.vaults = [BalancerV2Vault(vault, asynchronous=self.asynchronous) for vault in BALANCER_V2_VAULTS]

    @stuck_coro_debugger
    async def get_token_price(self, token_address: Address, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
        if deepest_pool := await self.deepest_pool_for(token_address, block=block, sync=False):
            return await deepest_pool.get_token_price(token_address, block, skip_cache=skip_cache, sync=False)
    
    # NOTE: we need a tiny semaphore here because balancer is super arduous and every unpricable token must pass thru this section
    @a_sync.Semaphore(10)
    @stuck_coro_debugger
    async def deepest_pool_for(self, token_address: Address, block: Optional[Block] = None) -> Optional[BalancerV2Pool]:
        kwargs = {"token_address": token_address, "block": block}
        deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs)
        if deepest_pools := {vault.address: deepest_pool async for vault, deepest_pool in deepest_pools if deepest_pool is not None}:
            logger.debug("%s deepest pools for %s at %s: %s", self, token_address, block, deepest_pools)
            async for pool in BalancerV2Pool.get_balance.map(deepest_pools.values(), **kwargs).keys(pop=True).aiterbyvalues(reverse=True):
                return pool
        
        # TODO: afilter
        # deepest_pools = BalancerV2Vault.deepest_pool_for.map(self.vaults, **kwargs).values(pop=True).afilter()
        # async for pool in BalancerV2Pool.get_balance.map(deepest_pools, **kwargs).keys(pop=True).aiterbyvalues(reverse=True):
        #     return pool

balancer = BalancerV2(asynchronous=True)

_lookup_balance_from_tuple: Callable[[Tuple[Any, T]], T] = lambda pool_and_balance: pool_and_balance[1]
"Takes a tuple[K, V] and returns V."

_warned = set()



import asyncio
from decimal import Decimal
from typing import Optional

import a_sync
from brownie.convert.datatypes import EthAddress
from multicall import Call
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.datatypes import Block, UsdPrice


@a_sync.a_sync(default='sync')
async def is_basketdao_index(address: EthAddress) -> bool:
    """
    Check if the given token is a BasketDAO token.

    Args:
        address: The address of the token to check.

    Returns:
        True if the token is a BasketDAO token, False otherwise.

    Example:
        >>> is_bd = is_basketdao("0x0ac58Df435D3dC9F6e079B2C5F358A4b7e861B69")
        >>> print(is_bd)
        True
    """
    try:
        return bool(await Call(address, 'getAssetsAndBalances()(address[],uint[])'))
    except (ContractLogicError, ValueError):
        return False

@a_sync.a_sync(default='sync')
async def get_price(
    address: EthAddress, 
    block: Optional[Block] = None, 
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    """
    Get the price of a BasketDAO token in USD.

    Args:
        address: The address of the BasketDAO token.
        block (optional): The block number to query. Defaults to None (latest).
        skip_cache (optional): Whether to bypass the disk cache. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the BasketDAO token in USD.

    Raises:
        PriceError: If unable to fetch the price.

    Example:
        >>> price = get_price("0x0ac58Df435D3dC9F6e079B2C5F358A4b7e861B69")
        >>> print(price)
        1.05
    """
    balances, total_supply = await asyncio.gather(
        Call(address, 'getAssetsAndBalances()(address[],uint[])',block_id=block),
        ERC20(address, asynchronous=True).total_supply_readable(block=block),
    )
    balances = (WeiBalance(balance, token, block, skip_cache=skip_cache) for token, balance in zip(balances[0], balances[1]))
    tvl = await WeiBalance.value_usd.sum(balances, sync=False)
    return UsdPrice(tvl / Decimal(total_supply))



import asyncio
import logging
from typing import Callable, List, Optional

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.convert.datatypes import EthAddress, HexString
from multicall import Call
from typing_extensions import Self

from y import convert
from y import ENVIRONMENT_VARIABLES as ENVS
from y.contracts import Contract, has_method
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.exceptions import UnsupportedNetwork, call_reverted
from y.networks import Network
from y.utils import a_sync_ttl_cache

try:
    from eth_abi import encode
    encode_bytes: Callable[[str], bytes] = lambda s: encode(["bytes32"], [s.encode()])
except ImportError:
    from eth_abi import encode_single
    encode_bytes: Callable[[str], bytes] = lambda s: encode_single("bytes32", s.encode())

logger = logging.getLogger(__name__)

addresses = {
    Network.Mainnet: '0x823bE81bbF96BEc0e25CA13170F5AaCb5B79ba83',
    Network.Optimism: '0x95A6a3f44a70172E7d50a9e28c85Dfd712756B8C',
}


class Synthetix(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = False) -> None:
        if chain.id not in addresses:
            raise UnsupportedNetwork("synthetix is not supported on this network")
        self.asynchronous = asynchronous
        super().__init__()
    
    @a_sync.aka.property
    async def address_resolver(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id])
    __address_resolver__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.a_sync(ram_cache_maxsize=256)
    async def get_address(self, name: str, block: Block = None) -> Contract:
        """
        Get contract from Synthetix registry.
        See also https://docs.synthetix.io/addresses/
        """
        address_resolver = await self.__address_resolver__
        address = await address_resolver.getAddress.coroutine(encode_bytes(name), block_identifier=block)
        proxy = await Contract.coroutine(address)
        return await Contract.coroutine(await proxy.target.coroutine(block_identifier=block)) if hasattr(proxy, 'target') else proxy

    @a_sync.aka.cached_property
    async def synths(self) -> List[EthAddress]:
        """
        Get target addresses of all synths.
        """
        proxy_erc20 = await self.get_address('ProxyERC20', sync=False)
        synths = await proxy_erc20.availableSynths.map(range(await proxy_erc20.availableSynthCount))
        logger.info('loaded %s synths', len(synths))
        return synths
        
    async def is_synth(self, token: AnyAddressType) -> bool:
        """returns `True` if a `token` is a synth, `False` if not"""
        token = convert.to_address(token)
        try:
            if await synthetix.get_currency_key(token, sync=False):
                return True
            if await has_method(token, 'target()(address)', sync=False):
                target = await Call(token, 'target()(address)')
                return target in await synthetix.synths and await Call(target, 'proxy()(address)') == token
            return False
        except Exception as e:
            if "invalid jump destination" in str(e):
                return False
            raise
    
    @a_sync_ttl_cache
    async def get_currency_key(self, token: AnyAddressType) -> Optional[HexString]:
        target = await Call(token, ['target()(address)']) if await has_method(token, 'target()(address)', sync=False) else token
        return await Call(target, ['currencyKey()(bytes32)']) if await has_method(token, 'currencyKey()(bytes32)', sync=False) else None
    
    async def get_price(self, token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
        """
        Get a price of a synth in dollars.
        """
        token = convert.to_address(token)
        rates, key = await asyncio.gather(
            self.get_address('ExchangeRates', block=block, sync=False),
            self.get_currency_key(token, sync=False)
        )
        if await rates.rateIsStale.coroutine(key, block_identifier=block):
            return None
        try:
            return UsdPrice(await rates.rateForCurrency.coroutine(key, block_identifier=block, decimals=18))
        except Exception as e:
            if not call_reverted(e):
                raise

synthetix = Synthetix(asynchronous=True) if chain.id in addresses else set()



import asyncio
import logging
import weakref
from types import MethodType
from typing import List, NoReturn, Optional, Tuple, TypeVar, Union

import a_sync
from brownie import chain
from lazy_logging import LazyLoggerFactory
from typing_extensions import ParamSpec

from y.datatypes import AnyAddressType, Block
from y.networks import Network

T = TypeVar('T')
P = ParamSpec('P')

yLazyLogger = LazyLoggerFactory("YPRICEMAGIC")

logger = logging.getLogger(__name__)

class PriceLogger(logging.Logger):
    enabled: bool
    address: str
    block: int
    key: Tuple[AnyAddressType, Block, Optional[str], str]
    debug_task: Optional["asyncio.Task[None]"]
    def close(self) -> None:
        ...

def get_price_logger(token_address: AnyAddressType, block: Block, *, symbol: str = None, extra: str = '', start_task: bool = False) -> PriceLogger:
    address = str(token_address)
    name = f"y.prices.{Network.label()}.{address}.{block}"
    if extra: 
        name += f".{extra}"
    # the built-in logging module caches loggers but we need to make sure they have the proper members for ypm
    if logger := _all_price_loggers.get(name, None):
        return logger
    logger = logging.getLogger(name)
    logger.address = address
    logger.block = block
    if logger.level != logger.parent.level:
        logger.setLevel(logger.parent.level)
    logger.enabled = logger.isEnabledFor(logging.DEBUG)
    if start_task and logger.enabled:
        # will kill itself when this logger is garbage collected
        logger.debug_task = a_sync.create_task(
            coro=_debug_tsk(symbol, weakref.ref(logger)), 
            name=f"_debug_tsk({symbol}, {logger})", 
            log_destroy_pending=False,
        )
    logger.close = MethodType(_close_logger, logger)
    _all_price_loggers[name] = logger
    return logger

def _close_logger(logger: PriceLogger) -> None:
    # since we make a lot of these we don't want logging module to cache them
    logger.debug("closing %s", logger)
    logging._lock.acquire()
    logging.Logger.manager.loggerDict.pop(logger.name, None)
    logging._lock.release()

async def _debug_tsk(symbol: Optional[str], logger_ref: "weakref.ref[logging.Logger]") -> NoReturn:
    """Prints a log every 1 minute until the creating coro returns"""
    if symbol:
        args = "price still fetching for %s", symbol
    else:
        args = "still fetching...",
    while True:
        await asyncio.sleep(60)
        logger = logger_ref()
        if logger is None:
            return
        logger.debug(*args)

_all_price_loggers: "weakref.WeakValueDictionary[str, PriceLogger]" = weakref.WeakValueDictionary()

def enable_debug_logging(logger: str = 'y') -> None:
    logger = logging.getLogger(logger)
    logger.setLevel(logging.DEBUG)
    if not logger.handlers:
        logger.addHandler(logging.StreamHandler())

NETWORK_DESCRIPTOR_FOR_ISSUE_REQ =f'name ({Network.name()})' if Network.name() else f'chainid ({chain.id})'

def _gh_issue_request(issue_request_details: Union[str, List[str]], _logger = None) -> None:

    if _logger is None: _logger = logger

    if type(issue_request_details) == str:
        _logger.warning(issue_request_details)

    elif type(issue_request_details) == list:
        for message in issue_request_details:
            _logger.warning(message)

    _logger.warning('Please create an issue and/or create a PR at https://github.com/BobTheBuidler/ypricemagic')
    _logger.warning(f'In your issue, please include the network {NETWORK_DESCRIPTOR_FOR_ISSUE_REQ} and the detail shown above.')
    _logger.warning('and I will add it soon :). This will not prevent ypricemagic from fetching price for this asset.')



import asyncio
import logging
import math
from collections import defaultdict
from functools import cached_property, lru_cache
from itertools import cycle
from typing import AsyncIterator, DefaultDict, List, Optional, Tuple

import a_sync
import eth_retry
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.network.event import _EventItem
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, ContractBase
from y.constants import usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import ContractNotVerified, TokenNotFound, UnsupportedNetwork, call_reverted
from y.interfaces.uniswap.quoterv3 import UNIV3_QUOTER_ABI
from y.networks import Network
from y.utils.events import ProcessedEvents

try:
    from eth_abi.packed import encode_packed
except ImportError:
    from eth_abi.packed import encode_abi_packed as encode_packed

# https://github.com/Uniswap/uniswap-v3-periphery/blob/main/deploys.md
UNISWAP_V3_FACTORY = '0x1F98431c8aD98523631AE4a59f267346ea31F984'
UNISWAP_V3_QUOTER = '0xb27308f9F90D607463bb33eA1BeBb41C27CE5AB6'

logger = logging.getLogger(__name__)

# same addresses on all networks
addresses = {
    Network.Mainnet: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Arbitrum: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000],
    },
    Network.Optimism: {
        'factory': UNISWAP_V3_FACTORY,
        'quoter': UNISWAP_V3_QUOTER,
        'fee_tiers': [3000, 500, 10_000, 100],
    },
    Network.Base: {
        'factory': '0x33128a8fC17869897dcE68Ed026d694621f6FDfD',
        'quoter': '0x3d4e44Eb1374240CE5F1B871ab261CD16335B76a', # quoter v2
        'fee_tiers': [3000, 500, 10_000, 100],
    }
}

FEE_DENOMINATOR = 1_000_000


class UniswapV3Pool(ContractBase):
    __slots__ = 'fee', 'token0', 'token1', 'tick_spacing'
    def __init__(
        self,
        address: Address,
        token0: Address, 
        token1: Address, 
        tick_spacing: int, 
        fee: int, 
        deploy_block: int,
        asynchronous: bool = False
    ) -> None:
        super().__init__(address, asynchronous=asynchronous)
        self.token0 = ERC20(token0, asynchronous=asynchronous)
        self.token1 = ERC20(token1, asynchronous=asynchronous)
        self.tick_spacing = tick_spacing
        self.fee = fee
        self._deploy_block = deploy_block

    def __contains__(self, token: Address) -> bool:
        return token in [self.token0, self.token1]
    
    def __getitem__(self, token: Address) -> ERC20:
        if token not in self:
            raise TokenNotFound(token, self)
        return ERC20(token, asynchronous=self.asynchronous)

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60, semaphore=10000)  # lets try a semaphore here
    async def check_liquidity(self, token: AnyAddressType, block: Block) -> Optional[int]:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block < await self.deploy_block(sync=False):
            logger.debug("block %s prior to %s deploy block", block, self)
            return 0
        try:
            liquidity = await self[token].balance_of(self.address, block, sync=False)
        except ContractNotVerified:
            logger.debug("%s is not verified and we cannot fetch balance the usual way. returning 0.", token)
            return 0
        logger.debug("%s liquidity for %s at %s: %s", self, token, block, liquidity)
        return liquidity
    
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def _check_liquidity_token_out(self, token_in: AnyAddressType, block: Block) -> Optional[int]:
        return await self.check_liquidity(self._get_token_out(token_in), block=block, sync=False)
    
    @lru_cache
    def _get_token_out(self, token_in: ERC20) -> ERC20:
        if token_in == self.token0:
            return self.token1
        elif token_in == self.token1:
            return self.token0
        raise TokenNotFound(token_in, self)


class UniswapV3(a_sync.ASyncGenericSingleton):
    def __init__(self, asynchronous: bool = True) -> None:
        self.asynchronous = asynchronous
        if chain.id not in addresses:
            raise UnsupportedNetwork('compound is not supported on this network')
        self.fee_tiers = addresses[chain.id]['fee_tiers']
        self.loading = False
        self._pools = {}

    def __contains__(self, asset) -> bool:
        return chain.id in addresses
    
    @cached_property
    def loaded(self) -> a_sync.Event:
        return a_sync.Event(name=self)

    @a_sync.aka.property
    async def factory(self) -> Contract:
        return await Contract.coroutine(addresses[chain.id]['factory'])
    __factory__: HiddenMethodDescriptor[Self, Contract]
    
    @a_sync.aka.cached_property
    async def quoter(self) -> Contract:
        quoter = addresses[chain.id]['quoter']
        try:
            return await Contract.coroutine(quoter)
        except ContractNotVerified:
            return Contract.from_abi("Quoter", quoter, UNIV3_QUOTER_ABI)
    __quoter__: HiddenMethodDescriptor[Self, Contract]

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV3Pool]:
        factory = await self.__factory__
        return UniV3Pools(factory, asynchronous=self.asynchronous)
    __pools__: HiddenMethodDescriptor[Self, "UniV3Pools"]

    async def pools_for_token(self, token: Address, block: Block) -> AsyncIterator[UniswapV3Pool]:
        pools = await self.__pools__
        async for pool in pools.objects(to_block=block):
            if token in pool:
                yield pool
    
    @stuck_coro_debugger
    @a_sync.a_sync(cache_type='memory', ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(
        self, 
        token: Address, 
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (), # unused
        skip_cache: bool = ENVS.SKIP_CACHE,  # unused
        ) -> Optional[UsdPrice]:

        quoter = await self.__quoter__
        if block and block < await contract_creation_block_async(quoter, True):
            return None

        paths = [[token, fee, usdc.address] for fee in self.fee_tiers]
        if token != weth:
            paths += [
                [token, fee, weth.address, self.fee_tiers[0], usdc.address] for fee in self.fee_tiers
            ]
        logger.debug("paths: %s", paths)
        
        amount_in = await ERC20(token, asynchronous=True).scale
        results = await asyncio.gather(*(self._quote_exact_input(path, amount_in, block) for path in paths), return_exceptions=True)

        for result in results:
            if isinstance(result, Exception) and not call_reverted(result):
                raise result.with_traceback(result.__traceback__)
            
        logger.debug("results: %s", results)
        
        outputs = [
            # Quoter v2 uses this weird return struct, we must unpack it to get amount out.
            (amount if isinstance(amount, int) else amount[0]) / _undo_fees(path) / 1e6
            for amount, path in zip(results, paths)
            if amount and not call_reverted(amount)
        ]
        logger.debug("outputs: %s", outputs)
        return UsdPrice(max(outputs)) if outputs else None

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60*60)
    async def check_liquidity(self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()) -> int:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if chain.id == Network.Mainnet and token == "0x6DEA81C8171D0bA574754EF6F8b412F2Ed88c54D":
            # LQTY, TODO refactor this somehow
            return 0
        
        quoter = await self.__quoter__
        if block and block < await contract_creation_block_async(quoter):
            logger.debug("block %s is before %s deploy block", block, quoter)
            return 0

        if token == weth.address:
            # NOTE: we need to filter these or else we will be fetching every pool
            #       for now, we only focus on weth/usdc pools
            filter_fn = lambda pool: pool._get_token_out(token) == usdc.address and pool not in ignore_pools
        else:
            filter_fn = lambda pool: pool not in ignore_pools

        token_out_tasks = UniswapV3Pool._check_liquidity_token_out.map(token_in=token, block=block)
        pools_for_token: a_sync.ASyncIterator[UniswapV3Pool] = self.pools_for_token(token, block)
        async for pool in pools_for_token.filter(filter_fn):
            # the mapping will start the tasks internally
            logger.debug("starting token_out_task for %s", pool)
            token_out_tasks[pool]
        if not token_out_tasks:
            return 0
        logger.debug("%s token_out_tasks for %s at %s: %s", self, token, block, token_out_tasks)
        
        # Since uni v3 liquidity can be provided asymmetrically, the most liquid pool in terms of `token` might not actually be the most liquid pool in terms of `token_out`
        # We need some spaghetticode here to account for these erroneous liquidity values
        # TODO: Refactor this
        token_out_liquidity: DefaultDict[ERC20, List[int]] = defaultdict(list)
        async for pool, liquidity in token_out_tasks.map(pop=False):
            logger.debug("%s liquidity for %s at %s: %s", pool, token, block, liquidity)
            token_out_liquidity[pool._get_token_out(token)].append(liquidity)
        logger.debug("%s token_out_liquidity: %s", token, token_out_liquidity)
        
        token_out_min_liquidity = {token_out: min(liquidities) for token_out, liquidities in token_out_liquidity.items()}

        token_in_tasks = UniswapV3Pool.check_liquidity.map(token=token, block=block)
        async for pool, liquidity in token_out_tasks.map(pop=True):
            token_out = pool._get_token_out(token)
            if len(token_out_liquidity[token_out]) > 1 and liquidity == token_out_min_liquidity[token_out]:
                logger.debug("ignoring liquidity for %s", pool)
            elif token_out == weth and liquidity < 10 ** 19:  # 10 ETH
                # NOTE: this is totally arbitrary, works for all known cases but eventually will probably cause issues
                logger.debug("insufficient liquidity for %s", pool)
            else:
                token_in_tasks[pool]

        liquidity = await token_in_tasks.max(pop=True, sync=False) if token_in_tasks else 0
        logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
        return liquidity
    
    @stuck_coro_debugger
    @eth_retry.auto_retry
    async def _quote_exact_input(self, path: List[list], amount_in: int, block: int) -> int:
        quoter = await self.__quoter__
        return await quoter.quoteExactInput.coroutine(_encode_path(path), amount_in, block_identifier=block)

def _encode_path(path) -> bytes:
    types = [type for _, type in zip(path, cycle(['address', 'uint24']))]
    return encode_packed(types, path)

def _undo_fees(path) -> float:
    fees = [1 - fee / FEE_DENOMINATOR for fee in path if isinstance(fee, int)]
    return math.prod(fees)


class UniV3Pools(ProcessedEvents[UniswapV3Pool]):
    __slots__ = "asynchronous", 
    def __init__(self, factory: Contract, asynchronous: bool = False):
        self.asynchronous = asynchronous
        super().__init__(addresses=[factory.address], topics=[factory.topics["PoolCreated"]])
    def _process_event(self, event: _EventItem) -> UniswapV3Pool:
        token0, token1, fee, tick_spacing, pool = event.values()
        return UniswapV3Pool(pool, token0, token1, fee, tick_spacing, event.block_number, asynchronous=self.asynchronous)
    def _get_block_for_obj(self, obj: UniswapV3Pool) -> int:
        return obj._deploy_block


try:
    uniswap_v3 = UniswapV3(asynchronous=True)
except UnsupportedNetwork:
    uniswap_v3 = None




import asyncio
import json
import logging
import threading
from collections import defaultdict
from typing import Any, Callable, Dict, List, Literal, NewType, Optional, Set, Tuple, Union

import a_sync
import dank_mids
import eth_retry
from brownie import ZERO_ADDRESS, chain, web3
from brownie.exceptions import CompilerError, ContractNotFound
from brownie.network.contract import (ContractEvents, _add_deployment,
                                      _ContractBase, _DeployedContractBase,
                                      _fetch_from_explorer, _resolve_address)
from brownie.typing import AccountsType
from checksum_dict import ChecksumAddressDict, ChecksumAddressSingletonMeta
from hexbytes import HexBytes
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert, exceptions
from y._decorators import stuck_coro_debugger
from y.datatypes import Address, AnyAddressType, Block
from y.interfaces.ERC20 import ERC20ABI
from y.networks import Network
from y.time import check_node, check_node_async
from y.utils.cache import memory
from y.utils.events import Events
from y.utils.gather import gather_methods

logger = logging.getLogger(__name__)

contract_threads = a_sync.PruningThreadPoolExecutor(16)

# cached Contract instance, saves about 20ms of init time
_contract_lock = threading.Lock()

# These tokens have trouble when resolving the implementation via the chain.
FORCE_IMPLEMENTATION = {
    Network.Mainnet: {
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48": "0xa2327a938Febf5FEC13baCFb16Ae10EcBc4cbDCF", # USDC as of 2022-08-10
        "0x3d1E5Cf16077F349e999d6b21A4f646e83Cd90c5": "0xf51fC5ae556F5B8c6dCf50f70167B81ceb02a2b2", # dETH as of 2024-02-15
    },
}.get(chain.id, {})


def Contract_erc20(address: AnyAddressType) -> "Contract":
    """
    Create a Contract instance for an ERC20 token.

    This function uses the standard ERC20 ABI instead of fetching the contract ABI from the block explorer.
    
    Args:
        address: The address of the ERC20 token.

    Returns:
        A Contract instance for the ERC20 token.
    """
    address = convert.to_address(address)
    return Contract.from_abi('ERC20',address,ERC20ABI)


def Contract_with_erc20_fallback(address: AnyAddressType) -> "Contract":
    """
    Create a Contract instance for an address, falling back to an ERC20 token if the contract is not verified.

    Args:
        address: The address of the contract or ERC20 token.

    Returns:
        A Contract instance for the contract address.
    """
    if isinstance(address, Contract):
        return address
    address = convert.to_address(address)
    try:
        return Contract(address)
    except exceptions.ContractNotVerified:
        return Contract_erc20(address)


@memory.cache()
#yLazyLogger(logger)
def contract_creation_block(address: AnyAddressType, when_no_history_return_0: bool = False) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.

    Args:
        address: The address of the contract.
        when_no_history_return_0: If True, return 0 when no history is found instead of raising a :class:`~exceptions.NodeNotSynced` exception. Default False.

    Returns:
        The block number at which the contract was created.

    Raises:
        exceptions.NodeNotSynced: If the node is not fully synced.
        ValueError: If the contract creation block cannot be determined.
    """
    address = convert.to_address(address)
    logger.debug("contract creation block %s", address)
    height = chain.height

    if height == 0:
        raise exceptions.NodeNotSynced(_NOT_SYNCED)
    
    check_node()

    lo, hi = 0, height
    barrier = 0
    warned = False
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        # TODO rewrite this so we can get deploy blocks for some contracts deployed on correct side of barrier
        try:
            if eth_retry.auto_retry(web3.eth.get_code)(address, mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if 'missing trie node' in str(e) and not warned:
                logger.warning('missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node.')
            elif 'Server error: account aurora does not exist while viewing' in str(e):
                if not warned:
                    logger.warning(str(e))
            elif 'No state available for block' in str(e):
                if not warned:
                    logger.warning(str(e))
            else:
                raise
            warned = True
            barrier = mid
            lo = mid
    if hi == lo + 1 == barrier + 1 and when_no_history_return_0:
        logger.warning('could not determine creation block for %s on %s (deployed prior to barrier)', address, Network.name())
        logger.debug("contract creation block %s -> 0", address)
        return 0
    if hi != height:
        logger.debug("contract creation block %s -> %s", address, hi)
        return hi
    raise ValueError(f"Unable to find deploy block for {address} on {Network.name()}")

get_code = eth_retry.auto_retry(dank_mids.eth.get_code)
creation_block_semaphore = a_sync.ThreadsafeSemaphore(10)

@a_sync.a_sync(cache_type='memory')
@stuck_coro_debugger
@eth_retry.auto_retry
async def contract_creation_block_async(address: AnyAddressType, when_no_history_return_0: bool = False) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.

    Args:
        address: The address of the contract.
        when_no_history_return_0: If True, return 0 when no history is found instead of raising a :class:`~exceptions.NodeNotSynced` exception. Default False.

    Returns:
        The block number at which the contract was created.

    Raises:
        exceptions.NodeNotSynced: If the node is not fully synced.
        ValueError: If the contract creation block cannot be determined.
    """
    from y._db.utils import contract as db

    address = convert.to_address(address)
    if deploy_block := await db.get_deploy_block(address):
        return deploy_block
    
    logger.debug(f"contract creation block {address}")
    height = await dank_mids.eth.block_number

    if height == 0:
        raise exceptions.NodeNotSynced(_NOT_SYNCED)
    
    await check_node_async()

    lo, hi = 0, height
    barrier = 0
    warned = False
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        # TODO rewrite this so we can get deploy blocks for some contracts deployed on correct side of barrier
        try:
            if await get_code(address, mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if 'missing trie node' in str(e):
                if not warned:
                    logger.warning('missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node.')
            elif 'Server error: account aurora does not exist while viewing' in str(e):
                if not warned:
                    logger.warning(str(e))
            elif 'No state available for block' in str(e):
                if not warned:
                    logger.warning(str(e))
            else:
                raise
            warned = True
            barrier = mid
            lo = mid
    if hi == lo + 1 == barrier + 1 and when_no_history_return_0:
        logger.warning(f'could not determine creation block for {address} on {Network.name()} (deployed prior to barrier)')
        logger.debug(f"contract creation block {address} -> 0")
        return 0
    if hi != height:
        logger.debug(f"contract creation block {address} -> {hi}")
        db.set_deploy_block(address, hi)
        return hi
    raise ValueError(f"Unable to find deploy block for {address} on {Network.name()}")


# this defaultdict prevents congestion in the contracts thread pool
address_semaphores = defaultdict(lambda: a_sync.ThreadsafeSemaphore(1))

class ContractEvents(ContractEvents):
    def __getattr__(self, name: str) -> Events:
        return super().__getattr__(name)
    
class Contract(dank_mids.Contract, metaclass=ChecksumAddressSingletonMeta):
    """
    Though it may look complicated, a ypricemagic Contract object is simply a brownie Contract object with a few modifications:
    1. Contracts will not be compiled. This allows you to more quickly fetch contracts from the block explorer and prevents you from having to download and install compilers.
    2. To each contract method, a `coroutine` property has been defined which allows you to make asynchronous calls using the following syntax:
    ```python
    Contract(0xAddress).methodName.coroutine(*args, block_identifier=12345678)
    ```
    3. A few attributes were removed in order to minimize the size of a Contract object in memory: 
        - ast, bytecode, coverageMap, deployedBytecode, deployedSourceMap, natspec, opcodes, pcMap
    4. There are a few new util methods but they're not officially supported yet and may change without warning:
        - has_method
        - has_methods
        - has_methods_async
        - build_name
        - build_name_async
        - get_code
    """
    # the default state for Contract objects
    verified = True

    events: ContractEvents
    _ChecksumAddressSingletonMeta__instances: ChecksumAddressDict["Contract"]

    @eth_retry.auto_retry
    def __init__(
        self, 
        address: AnyAddressType, 
        owner: Optional[AccountsType] = None, 
        require_success: bool = True, 
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> None:
        
        address = str(address)
        if address.lower() in ["0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee", ZERO_ADDRESS]:
            raise ContractNotFound(f"{address} is not a contract.")
        if require_success and address in _unverified:
            raise exceptions.ContractNotVerified(address)

        with _contract_lock:
            # autofetch-sources: false
            # Try to fetch the contract from the local sqlite db.
            try:
                super().__init__(address, owner=owner)
                if not isinstance(self.verified, bool) and self.verified is not None:
                    logger.warning(f'`Contract("{address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.')
            except AssertionError as e:
                raise CompilerError("y.Contract objects work best when we bypass compilers. In this case, it will *only* work when we bypass. Please ensure autofetch_sources=False in your brownie-config.yaml and rerun your script.") from None
            except IndexError as e:
                if str(e) == "pop from an empty deque":
                    raise CompilerError("y.Contract objects work best when we bypass compilers. In this case, it will *only* work when we bypass. Please ensure autofetch_sources=False in your brownie-config.yaml and rerun your script.") from None
                raise
            except ValueError as e:
                if not str(e).startswith("Unknown contract address: "):
                    raise
                logger.debug(f"{e}")
                try:                  
                    name, abi = _resolve_proxy(address)
                    build = {"abi": abi, "address": address, "contractName": name, "type": "contract"}
                    self.__init_from_abi__(build, owner=owner, persist=True)
                except exceptions.InvalidAPIKeyError:
                    # re-raise with a cleaner traceback
                    raise exceptions.InvalidAPIKeyError from None
                except (ContractNotFound, exceptions.ContractNotVerified) as e:
                    if isinstance(e, exceptions.ContractNotVerified):
                        _unverified.add(address)
                    if require_success:
                        raise
                    try:
                        if isinstance(e, exceptions.ContractNotVerified):
                            self.verified = False
                            self._build = {"contractName": "Non-Verified Contract"}
                        else:
                            self.verified = None
                            self._build = {"contractName": "Broken Contract"}
                    except AttributeError:
                        logger.warning(f'`Contract("{address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.')
                # Patch the Contract with coroutines for each method.
                dank_mids.patch_contract(self)

        if self.verified:
            _setup_events(self)         # Init an event container for each topic
            _squeeze(self)              # Get rid of unnecessary memory-hog properties

            self._ttl_cache_popper: Union[Literal["disabled"], int, asyncio.TimerHandle]
            try:
                self._ttl_cache_popper = "disabled" if cache_ttl is None else asyncio.get_running_loop().call_later(cache_ttl, self._ChecksumAddressSingletonMeta__instances.pop, self.address, None)
            except RuntimeError:
                self._ttl_cache_popper = cache_ttl

    @classmethod
    @a_sync.a_sync
    def from_abi(
        cls,
        name: str,
        address: str,
        abi: List,
        owner: Optional[AccountsType] = None,
        persist: bool = True,
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:
        """
        Create a Contract instance from an ABI.

        Args:
            name: The name of the contract.
            address: The address of the contract.
            abi: The ABI of the contract.
            owner (optional): The owner of the contract. Default None.
            persist (optional): If True, persist the contract in brownie's local contract database. Default True.
            cache_ttl (optional): The time-to-live for the contract cache in seconds. Default set in :mod:`~y.ENVIRONMENT_VARIABLES`.

        Returns:
            A Contract instance for the given ABI.
        """
        self = cls.__new__(cls)
        build = {"abi": abi, "address": _resolve_address(address), "contractName": name, "type": "contract"}
        self.__init_from_abi__(build, owner, persist)
        dank_mids.patch_contract(self)  # Patch the Contract with coroutines for each method.
        _setup_events(self)             # Init an event container for each topic
        _squeeze(self)                  # Get rid of unnecessary memory-hog properties
        try:
            self._ttl_cache_popper = "disabled" if cache_ttl is None else asyncio.get_running_loop().call_later(cache_ttl, cls._ChecksumAddressSingletonMeta__instances.pop, self.address, None)
        except RuntimeError:
            self._ttl_cache_popper = cache_ttl
        return self

    @classmethod
    async def coroutine(
        cls, 
        address: AnyAddressType, 
        require_success: bool = True, 
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:
        
        address = str(address)
        if contract := cls._ChecksumAddressSingletonMeta__instances.get(address, None):
            return contract
        # dict lookups faster than string comparisons, keep this behind the singleton check
        if address.lower() in ["0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee", ZERO_ADDRESS]:
            raise ContractNotFound(f"{address} is not a contract.") from None
        try:
            # We do this so we don't clog the threadpool with multiple jobs for the same contract.
            return await _contract_queue(address, require_success=require_success, cache_ttl=cache_ttl)
        except (ContractNotFound, exceptions._ExplorerError, CompilerError) as e:
            # re-raise with nicer traceback
            raise type(e)(*e.args) from None
    
    @classmethod
    @stuck_coro_debugger
    async def _coroutine(
        cls, 
        address: AnyAddressType, 
        require_success: bool = True, 
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:
        """
        Internal method to create a Contract instance asynchronously.

        Args:
            address: The address of the contract.
            require_success: If True, raise an exception if the contract cannot be initialized.
            cache_ttl: The time-to-live for the contract cache in seconds.

        Returns:
            A Contract instance for the given address.
        """
        contract = await contract_threads.run(cls, address, require_success=require_success)

        if not contract.verified or contract._ttl_cache_popper == "disabled":
            pass
    
        elif cache_ttl is None:
            if isinstance(contract._ttl_cache_popper, asyncio.TimerHandle):
                contract._ttl_cache_popper.cancel()
            contract._ttl_cache_popper = "disabled"
    
        elif isinstance(contract._ttl_cache_popper, int):
            cache_ttl = max(contract._ttl_cache_popper, cache_ttl)
            contract._ttl_cache_popper = asyncio.get_running_loop().call_later(cache_ttl, cls._ChecksumAddressSingletonMeta__instances.pop, contract.address, None)

        elif asyncio.get_running_loop().time() + cache_ttl > contract._ttl_cache_popper.when():
            contract._ttl_cache_popper.cancel()
            contract._ttl_cache_popper = asyncio.get_running_loop().call_later(cache_ttl, cls._ChecksumAddressSingletonMeta__instances.pop, contract.address, None)
        return contract

    def __init_from_abi__(self, build: Dict, owner: Optional[AccountsType] = None, persist: bool = True) -> None:
        """
        Initialize a Contract instance from an ABI.

        Args:
            build: The build information for the contract.
            owner (optional): The owner of the contract. Default None.
            persist (optional): If True, persist the contract in the local database. Default True.

        Returns:
            The initialized Contract instance.
        """
        _ContractBase.__init__(self, None, build, {})  # type: ignore
        _DeployedContractBase.__init__(self, build['address'], owner, None)
        if persist:
            _add_deployment(self)
        try:
            self.verified = True
        except AttributeError:
            logger.warning(f'`Contract("{self.address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.')
        return self

    def has_method(self, method: str, return_response: bool = False) -> Union[bool,Any]:
        """
        Check if the contract has a specific method.

        Args:
            method: The name of the method to check for.
            return_response (optional): If True, return the response of the method call instead of a boolean. Default False.

        Returns:
            A boolean indicating whether the contract has the method, or the response of the method call if return_response is True.
        """
        return has_method(self.address, method, return_response=return_response, sync=False)

    async def has_methods(
        self, 
        methods: List[str],
        _func: Union[any, all] = all
    ) -> bool:
        """
        Check if the contract has all the specified methods.

        Args:
            methods: A list of method names to check for.
            _func (optional): The function to use for combining the results (either :func:`all` or :func:`any`). Default :func:`all`.

        Returns:
            A boolean indicating whether the contract has all the specified methods.
        """
        return await has_methods(self.address, methods, _func, sync=False)

    async def build_name(self, return_None_on_failure: bool = False) -> Optional[str]:
        """
        Get the build name of the contract.

        Args:
            return_None_on_failure (optional): If True, return None if the build name cannot be determined instead of raising an Exception. Default False.

        Returns:
            The build name of the contract, or None if the build name cannot be determined and return_None_on_failure is True.
        """
        return await build_name(self.address, return_None_on_failure=return_None_on_failure, sync=False)
    
    async def get_code(self, block: Optional[Block] = None) -> HexBytes:
        """
        Get the bytecode of the contract at a specific block.

        Args:
            block (optional): The block number at which to get the bytecode. Defaults to latest block.

        Returns:
            The bytecode of the contract at the specified block.
        """
        return await get_code(self.address, block=block)

_contract_queue = a_sync.SmartProcessingQueue(Contract._coroutine, num_workers=32)

@memory.cache()
# TODO: async this and put it into ydb for quicker startups
#yLazyLogger(logger)
def is_contract(address: AnyAddressType) -> bool:
    '''
    Checks to see if the input address is a contract. Returns `False` if:
    - The address is not and has never been a contract
    - The address used to be a contract but has self-destructed
    '''
    address = convert.to_address(address)
    return web3.eth.get_code(address) not in ['0x',b'']

@a_sync.a_sync(default='sync', cache_type='memory')
async def has_method(address: Address, method: str, return_response: bool = False) -> Union[bool,Any]:
    '''
    Checks to see if a contract has a `method` view method with no inputs.
    `return_response=True` will return `response` in bytes if `response` else `False`
    '''
    address = convert.to_address(address)
    try:
        response = await Call(address, [method])
        return False if response is None else response if return_response else True
    except Exception as e:
        if isinstance(e, ContractLogicError) or exceptions.call_reverted(e):
            return False
        raise

@stuck_coro_debugger
@a_sync.a_sync(default='sync', cache_type='memory', ram_cache_ttl=15*60)
async def has_methods(
    address: AnyAddressType, 
    methods: Tuple[str],
    _func: Callable = all # Union[any, all]
) -> bool:
    '''
    Checks to see if a contract has each view method (with no inputs) in `methods`.
    Pass `at_least_one=True` to only verify a contract has at least one of the methods.
    '''

    assert _func in [all, any], '`_func` must be either `any` or `all`'

    address = convert.to_address(address)
    try:
        return _func([result is not None for result in await gather_methods(address, methods)])
    except Exception as e:
        if not isinstance(e, ContractLogicError) and not exceptions.call_reverted(e): raise # and not out_of_gas(e): raise
        # Out of gas error implies one or more method is state-changing.
        # If `_func == all` we return False because `has_methods` is only supposed to work for public view methods with no inputs
        # If `_func == any` maybe one of the methods will work without "out of gas" error
        return False if _func == all else any(await asyncio.gather(*[has_method(address, method, sync=False) for method in methods]))


#yLazyLogger(logger)
@stuck_coro_debugger
async def probe(
    address: AnyAddressType, 
    methods: List[str],
    block: Optional[Block] = None,
    return_method: bool = False
) -> Any:
    address = convert.to_address(address)
    results = await gather_methods(address, methods, block=block, return_exceptions=True)
    logger.debug('probe results: %s', results)
    results = [(method, result) for method, result in zip(methods, results) if not isinstance(result, Exception) and result is not None]
    if len(results) not in [1,0]:
        logger.debug('multiple results: %s', results)
        if len(results) != 2 or results[0][1] != results[1][1]:
            raise AssertionError(f'`probe` returned multiple results for {address}: {results}. Must debug')
        method = results[0][0], results[1][0]
        result = results[0][1]
        results = [(method, result)]
        logger.debug('final results: %s', results)
    method, result = results[0] if len(results) == 1 else (None, None)
    if method:
        assert result is not None
    return (method, result) if return_method else result
    

@a_sync.a_sync(default='sync')
@stuck_coro_debugger
async def build_name(address: AnyAddressType, return_None_on_failure: bool = False) -> str:
    """
    Get the build name of a contract.

    Args:
        address: The address of the contract.
        return_None_on_failure (optional): If True, return None if the build name cannot be determined instead of raising an Exception. Default False.

    Returns:
        The build name of the contract, or None if the build name cannot be determined and return_None_on_failure is True.
    """
    try:
        contract = await Contract.coroutine(address)
        return contract.__dict__['_build']['contractName']
    except exceptions.ContractNotVerified:
        if not return_None_on_failure:
            raise
        return None

async def proxy_implementation(address: AnyAddressType, block: Optional[Block]) -> Address:
    """
    Get the implementation address for a proxy contract.

    Args:
        address: The address of the proxy contract.
        block (optional): The block number at which to get the implementation address. Defaults to latest block.

    Returns:
        The address of the implementation contract.
    """
    return await probe(address, ['implementation()(address)','target()(address)'], block)

def _squeeze(contract: Contract) -> Contract:
    """
    Reduce the contract size in RAM by removing large data structures from the build.

    Args:
        contract: The contract object to squeeze.

    Returns:
        The squeezed contract object.
    """
    for k in ["ast", "bytecode", "coverageMap", "deployedBytecode", "deployedSourceMap", "natspec", "opcodes", "pcMap"]:
        if contract._build and k in contract._build.keys():
            contract._build[k] = {}
    return contract

@eth_retry.auto_retry
def _extract_abi_data(address: Address):
    """
    Extract ABI data for a contract from the blockchain explorer.

    Args:
        address: The address of the contract.

    Returns:
        A tuple containing the contract name, ABI, and implementation address (if applicable).

    Raises:
        Various exceptions based on the API response and contract status.
    """
    try:
        data = _fetch_from_explorer(address, "getsourcecode", False)["result"][0]
    except ConnectionError as e:
        if '{"message":"Something went wrong.","result":null,"status":"0"}' in str(e):
            if chain.id == Network.xDai:
                raise ValueError('Rate limited by Blockscout. Please try again.') from e
            if web3.eth.get_code(address):
                raise exceptions.ContractNotVerified(address) from e
            else:
                raise ContractNotFound(address) from e
        raise
    except ValueError as e:
        if str(e).startswith("Failed to retrieve data from API") and "invalid api key" in str(e).lower():
            raise exceptions.InvalidAPIKeyError from e
        if exceptions.contract_not_verified(e):
            raise exceptions.ContractNotVerified(f'{address} on {Network.printable()}') from e
        elif "Unknown contract address:" in str(e):
            exc_type = exceptions.ContractNotVerified if is_contract(address) else ContractNotFound
            raise exc_type(str(e)) from e
        else:
            raise

    is_verified = bool(data.get("SourceCode"))
    if not is_verified:
        raise exceptions.ContractNotVerified(f"Contract source code not verified: {address}") from None
    name = data["ContractName"]
    abi = json.loads(data["ABI"])
    implementation = data.get("Implementation")
    return name, abi, implementation

def _resolve_proxy(address) -> Tuple[str, List]:
    """
    Resolve the implementation address for a proxy contract.

    Args:
        address: The address of the proxy contract.

    Returns:
        A tuple containing the contract name and ABI.
    """
    address = convert.to_address(address)
    name, abi, implementation = _extract_abi_data(address)
    as_proxy_for = None

    if address in FORCE_IMPLEMENTATION:
        implementation = FORCE_IMPLEMENTATION[address]
        name, implementation_abi, _ = _extract_abi_data(implementation)
        # Here we merge the proxy ABI with the implementation ABI
        # without doing this, we'd only get the implementation
        # and would lack any valid methods/events from the proxy itself.
        # Credit: Wavey@Yearn
        abi += implementation_abi
        return name, abi

    # always check for an EIP1967 proxy - https://eips.ethereum.org/EIPS/eip-1967
    implementation_eip1967 = web3.eth.get_storage_at(
        address, int(web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
    )
    # always check for an EIP1822 proxy - https://eips.ethereum.org/EIPS/eip-1822
    implementation_eip1822 = web3.eth.get_storage_at(address, web3.keccak(text="PROXIABLE"))

    # Just leave this code where it is for a helpful debugger as needed.
    if address == "":
        raise Exception(
            f"""implementation: {implementation}
            implementation_eip1967: {len(implementation_eip1967)} {implementation_eip1967}
            implementation_eip1822: {len(implementation_eip1822)} {implementation_eip1822}""")

    if len(implementation_eip1967) > 0 and int(implementation_eip1967.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1967[-20:])
    elif len(implementation_eip1822) > 0 and int(implementation_eip1822.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1822[-20:])
    elif implementation:
        # for other proxy patterns, we only check if etherscan indicates
        # the contract is a proxy. otherwise we could have a false positive
        # if there is an `implementation` method on a regular contract.
        try:
            # first try to call `implementation` per EIP897
            # https://eips.ethereum.org/EIPS/eip-897
            c = Contract.from_abi(name, address, abi)
            as_proxy_for = c.implementation.call()
        except Exception:
            # if that fails, fall back to the address provided by etherscan
            as_proxy_for = _resolve_address(implementation)

    if as_proxy_for:
        name, abi, _ = _extract_abi_data(as_proxy_for)
    return name, abi

def _setup_events(contract: Contract) -> None:
    """
    Helper function used to init contract event containers on a newly created `y.Contract` object.

    Args:
        contract: The contract object to set up events for.
    """
    if not hasattr(contract, 'events'):
        contract.events = ContractEvents(contract)
    for k, v in contract.topics.items():
        setattr(contract.events, k, Events(addresses=[contract.address], topics=[[v]]))

_Address = NewType("_Address", str)
_unverified: Set[_Address] = set()
"""A collection of unverified addresses that is used to prevent repetitive etherscan api calls"""


_NOT_SYNCED = "`chain.height` returns 0 on your node, which means it is not fully synced.\nYou can only use this function on a fully synced node."




from tests.fixtures import mainnet_only
from y.prices import magic

@mainnet_only
def test_piedao_get_price():
    token = "0x9A48BD0EC040ea4f1D3147C025cd4076A2e71e3e"
    assert magic.get_price(token, 15_000_000, skip_cache=True) == 1.0000156215170668


"""
Utility functions for performing various checks.
"""

from typing import Any, Iterable


def hasall(obj: Any, attrs: Iterable[str]) -> bool:
    """
    Check if an object has all the specified attributes.

    Args:
        obj: The object to check.
        attrs: An iterable of attribute names to check for.

    Returns:
        True if the object has all the specified attributes, False otherwise.

    Example:
        >>> class TestClass:
        ...     attr1 = 1
        ...     attr2 = 2
        >>> test_obj = TestClass()
        >>> hasall(test_obj, ['attr1', 'attr2'])
        True
        >>> hasall(test_obj, ['attr1', 'attr3'])
        False
    """
    return all(hasattr(obj, attr) for attr in attrs)



from typing import TYPE_CHECKING, Union

from brownie import Contract
from brownie.convert.datatypes import EthAddress, HexBytes
from eth_typing import AnyAddress, BlockNumber

if TYPE_CHECKING:
    from y.prices.dex.uniswap.v2 import UniswapV2Pool
    from y.prices.stable_swap.curve import CurvePool


Address = Union[str, HexBytes, AnyAddress, EthAddress]
"""
A union of types used to represent Ethereum addresses.
"""

Block = Union[int, BlockNumber]
"""
A union of types used to represent block numbers as integers.
"""

AddressOrContract = Union[Address,Contract]

AnyAddressType = Union[Address, Contract, int]
"""
A type alias representing any valid representation of an Ethereum address.
This can be an Address, a :class:`~y.Contract`, or an integer.
"""

Pool = Union["UniswapV2Pool", "CurvePool"]
"""
A union of types representing liquidity pools.
"""

class UsdValue(float):
    """
    Represents a USD value with custom string representation.
    """
    
    def __str__(self) -> str:
        """
        Return a string representation of the USD value.

        Returns:
            A string formatted as a USD value with 8 decimal places.
        """
        return f'${self:.8f}'

class UsdPrice(UsdValue):
    """
    Represents a USD price.
    """




from y.prices.stable_swap import belt, ellipsis, froyo, mstablefeederpool, saddle
from y.prices.stable_swap.curve import curve

__all__ = [
    "belt",
    "curve",
    "ellipsis",
    "froyo",
    "mstablefeederpool",
    "saddle",
]



import asyncio
import logging
from typing import Optional, Tuple

import a_sync
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import call_reverted
from y.utils import gather_methods
from y.utils.cache import optional_async_diskcache

_RESERVES_METHODS = 'token0()(address)', 'token1()(address)', 'usersAmounts()((uint,uint))'

logger = logging.getLogger(__name__)


@a_sync.a_sync(default='sync', cache_type='memory', ram_cache_ttl=5*60)
@optional_async_diskcache
async def is_popsicle_lp(token_address: AnyAddressType) -> bool:
    """
    Determines if the given token address is a Popsicle Finance LP token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Popsicle LP token, False otherwise.

    Example:
        >>> is_popsicle = is_popsicle_lp("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf")
        >>> print(is_popsicle)
        True
    """
    # NOTE: contract to check for reference (mainnet): 0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf
    return await has_methods(token_address, ('token0()(address)','token1()(address)','usersAmounts()((uint,uint))'), sync=False)

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None, *, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
    """
    Calculates the price of a Popsicle Finance LP token.

    Args:
        token: The address of the Popsicle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the LP token in USD, or None if the price cannot be determined.

    Example:
        >>> price = get_price("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Popsicle LP token in USD
    """
    address = convert.to_address(token)
    total_val = await get_tvl(address, block, skip_cache=skip_cache, sync=False)
    if total_val is None:
        return None
    total_supply = await ERC20(address, asynchronous=True).total_supply_readable(block, sync=False)
    return UsdPrice(total_val / total_supply)

@a_sync.a_sync(default='sync')
async def get_tvl(token: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdValue]:
    balances: Tuple[WeiBalance, WeiBalance]
    balances = await get_balances(token, block, skip_cache=skip_cache, _async_balance_objects=True, sync=False)
    return UsdValue(await WeiBalance.value_usd.sum(balances)) if balances else None

@a_sync.a_sync(default='sync')
async def get_balances(token: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE, _async_balance_objects: bool = False) -> Optional[Tuple[WeiBalance,WeiBalance]]:
    try:
        token0, token1, (balance0, balance1) = await gather_methods(convert.to_address(token), _RESERVES_METHODS, block=block)
    except Exception as e:
        if call_reverted(e):
            return None
        elif str(e) == "not enough values to unpack (expected 3, got 2)":
            # TODO determine if this is regular behavior when no tvl in pool or if this is bug to fix
            return None
        raise
    balance0 = WeiBalance(balance0, token0, block=block, skip_cache=skip_cache, asynchronous=_async_balance_objects)
    balance1 = WeiBalance(balance1, token1, block=block, skip_cache=skip_cache, asynchronous=_async_balance_objects)
    return balance0, balance1




import asyncio
import functools
import inspect
import logging
import time
from typing import AsyncIterator, Awaitable, Callable, NoReturn, TypeVar, Union, overload

from a_sync import ASyncGenericBase
from a_sync.a_sync.method import ASyncBoundMethod
from a_sync.iter import ASyncGeneratorFunction
from typing_extensions import Concatenate, ParamSpec

P = ParamSpec('P')
T = TypeVar('T')

def continue_on_revert(func: Callable[P, T]) -> Callable[P, T]:
    '''
    Decorates a call-making function. If the call reverts, returns `None`. Raises any other exceptions.
    '''
    from y.exceptions import continue_if_call_reverted

    if asyncio.iscoroutinefunction(func):
        @functools.wraps(func)
        async def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return await func(*args,**kwargs)
            except Exception as e:
                continue_if_call_reverted(e)
    elif callable(func):
        @functools.wraps(func)
        def continue_on_revert_wrap(*args: P.args, **kwargs: P.kwargs) -> Union[T, None]:
            try:
                return func(*args,**kwargs)
            except Exception as e:
                continue_if_call_reverted(e)
    else:
        raise NotImplementedError(f"Unable to decorate {func}")
    return continue_on_revert_wrap

B = TypeVar("B", bound=ASyncGenericBase)

@overload
def stuck_coro_debugger(fn: Callable[Concatenate[B, P], AsyncIterator[T]]) -> ASyncGeneratorFunction[P, T]:...
@overload
def stuck_coro_debugger(fn: Callable[Concatenate[B, P], Awaitable[T]]) -> ASyncBoundMethod[B, P, T]:...
@overload
def stuck_coro_debugger(fn: Callable[Concatenate[B, P], T]) -> ASyncBoundMethod[B, P, T]:...
@overload
def stuck_coro_debugger(fn: Callable[P, AsyncIterator[T]]) -> Callable[P, AsyncIterator[T]]:...
@overload
def stuck_coro_debugger(fn: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:...
def stuck_coro_debugger(fn):
    logger = logging.getLogger("y.stuck?")
    if inspect.isasyncgenfunction(fn):
        @functools.wraps(fn)
        async def stuck_async_gen_wrap(*args: P.args, **kwargs: P.kwargs) -> AsyncIterator[T]:
            if not logger.isEnabledFor(logging.DEBUG):
                async for thing in fn(*args, **kwargs):
                    yield thing
            t = asyncio.create_task(coro=_stuck_debug_task(logger, fn, args, kwargs), name="_stuck_debug_task")
            try:
                async for thing in fn(*args, **kwargs):
                    yield thing
                t.cancel()
                return
            except Exception as e:
                t.cancel()
                raise
        return stuck_async_gen_wrap
    else:
        @functools.wraps(fn)
        async def stuck_coro_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            if not logger.isEnabledFor(logging.DEBUG):
                return await fn(*args, **kwargs)
            t = asyncio.create_task(coro=_stuck_debug_task(logger, fn, args, kwargs), name="_stuck_debug_task")
            try:
                retval = await fn(*args, **kwargs)
                t.cancel()
            except Exception as e:
                t.cancel()
                raise
            return retval
        return stuck_coro_wrap

async def _stuck_debug_task(logger: logging.Logger, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> NoReturn:
    start = time.time()
    while True:
        await asyncio.sleep(300)
        logger._log(
            logging.DEBUG, 
            f"{fn.__module__}.{fn.__name__} still executing after {round((time.time() - start)/60, 2)}m with"
            + f" args {tuple(str(arg) for arg in args)}"
            + f" kwargs {dict((k, str(v)) for k, v in kwargs.items())}",
            (),
        )



from typing import Optional

import a_sync
from brownie import chain
from brownie.convert.datatypes import EthAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import Block, UsdPrice
from y.networks import Network
from y.prices import magic

MAPPING = {
    Network.Mainnet: {
        "0x4da27a545c0c5B758a6BA100e3a049001de870f5": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9", # stkaave -> aave
        "0x27D22A7648e955E510a40bDb058333E9190d12D4": "0x0cec1a9154ff802e7934fc916ed7ca50bde6844e", # ppool -> pool
    }
}.get(chain.id,{})

def is_one_to_one_token(token_address: EthAddress) -> bool:
    return token_address in MAPPING

@a_sync.a_sync(default='sync')
async def get_price(token_address: EthAddress, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
    return await magic.get_price(MAPPING[token_address], block=block, skip_cache=skip_cache, sync=False)




import pytest
from y.constants import STABLECOINS
from y.prices import magic

from tests.fixtures import blocks_for_contract

STABLECOINS = STABLECOINS.keys()

# TODO Implement this after imlementing non-stable stables to validate tokens' spots in `STABLECOINS`.
@pytest.mark.parametrize('token',STABLECOINS)
def test_stablecoins(token):
    for block in blocks_for_contract(token,20):
        # NOTE Placeholder.
        assert magic.get_price(token, block, skip_cache=True) == 1, 'Stablecoin price not $1'



import asyncio
from typing import Optional, Tuple

import a_sync
from brownie.exceptions import ContractNotFound

from y import ENVIRONMENT_VARIABLES as ENVS
from y import Contract
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20, WeiBalance
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import MessedUpBrownieContract
from y.utils import gather_methods, hasall

_CHECK_METHODS = 'getReserves','token0','token1'
_TOKEN_METHODS = 'token0()(address)', 'token1()(address)'

async def is_generic_amm(lp_token_address: AnyAddressType) -> bool:
    """
    Check if the given address is a generic AMM LP token.

    Args:
        lp_token_address: The address to check.

    Returns:
        True if the address is a generic AMM LP token, False otherwise.

    Example:
        >>> address = "0x6B175474E89094C44Da98b954EedeAC495271d0F"
        >>> is_amm = await is_generic_amm(address)
        >>> print(is_amm)
        False
    """
    try:
        contract = await Contract.coroutine(lp_token_address, require_success=False)
        return contract.verified and hasall(contract, _CHECK_METHODS)
    except ContractNotFound:
        return False
    except MessedUpBrownieContract:
        # probably false, can get more specific when there's a need.
        return False
        
class GenericAmm(a_sync.ASyncGenericBase):
    """
    A class for handling generic Automated Market Maker (AMM) Liquidity Pool (LP) tokens.

    This class provides methods to interact with and price generic AMM LP token contracts.
    """

    def __init__(self, asynchronous: bool = False) -> None:
        """
        Initialize the GenericAmm instance.

        Args:
            asynchronous (optional): Whether methods will return coroutines by default. Defaults to False.
        """
        self.asynchronous = asynchronous
    
    @stuck_coro_debugger
    async def get_price(self, lp_token_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
        """
        Get the price of the LP token in USD.

        Args:
            lp_token_address: The address of the LP token.
            block (optional): The block number to query. Defaults to None (latest).
            skip_cache (optional): Whether to skip cache. Defaults to :obj:`ENVS.SKIP_CACHE`.

        Returns:
            The price of the LP token in USD.

        Example:
            >>> amm = GenericAmm(asynchronous=True)
            >>> address = "0x6B175474E89094C44Da98b954EedeAC495271d0F"
            >>> price = await amm.get_price(address)
            >>> print(price)
            1.0
        """
        tvl, total_supply = await asyncio.gather(
            self.get_tvl(lp_token_address, block=block, skip_cache=skip_cache, sync=False),
            ERC20(lp_token_address, asynchronous=True).total_supply_readable(block=block),
        )
        if total_supply is None:
            return None
        elif total_supply == 0:
            return 0
        return UsdPrice(tvl / total_supply)
    
    @stuck_coro_debugger
    @a_sync.a_sync(cache_type='memory')
    async def get_tokens(self, lp_token_address: AnyAddressType) -> Tuple[ERC20,ERC20]:
        """
        Get the tokens in the AMM pool.

        Args:
            lp_token_address: The address of the LP token.

        Returns:
            A tuple containing the two ERC20 tokens in the pool.

        Example:
            >>> amm = GenericAmm(asynchronous=True)
            >>> address = "0x6B175474E89094C44Da98b954EedeAC495271d0F"
            >>> tokens = await amm.get_tokens(address)
            >>> print(tokens)
            (ERC20(0x123...), ERC20(0x456...))
        """
        tokens = await gather_methods(lp_token_address, _TOKEN_METHODS)
        return tuple(ERC20(token, asynchronous=self.asynchronous) for token in tokens)
    
    @stuck_coro_debugger
    async def get_tvl(self, lp_token_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdValue:
        """
        Get the Total Value Locked (TVL) in the AMM pool.

        Args:
            lp_token_address: The address of the LP token.
            block (optional): The block number to query. Defaults to None (latest).
            skip_cache (optional): Whether to skip cache. Defaults to :obj:`ENVS.SKIP_CACHE`.

        Returns:
            The Total Value Locked in USD.

        Example:
            >>> amm = GenericAmm(asynchronous=True)
            >>> address = "0x6B175474E89094C44Da98b954EedeAC495271d0F"
            >>> tvl = await amm.get_tvl(address)
            >>> print(tvl)
            1000000.0
        """
        lp_token_contract = await Contract.coroutine(lp_token_address)
        tokens, reserves = await asyncio.gather(
            self.get_tokens(lp_token_address, sync=False),
            lp_token_contract.getReserves.coroutine(block_identifier=block),
        )
        reserves = (WeiBalance(reserve, token, block=block, skip_cache=skip_cache) for token, reserve in zip(tokens,reserves))
        return UsdValue(await WeiBalance.value_usd.sum(reserves, sync=False))


generic_amm = GenericAmm(asynchronous=True)
"""A global instance of :class:`~GenericAmm` with asynchronous mode enabled."""



import logging
from typing import TYPE_CHECKING, Any, Optional

from brownie import Contract as BrownieContract
from brownie.exceptions import CompilerError
from web3.exceptions import ContractLogicError

from y.datatypes import AnyAddressType

if TYPE_CHECKING:
    from y.prices.dex.uniswap.v2 import UniswapV2Pool

logger = logging.getLogger(__name__)

# General

class yPriceMagicError(ValueError):
    """
    Custom exception for ypricemagic-related errors.

    This exception is raised when any error occurs inside of ypricemagic.
    For example, if a price fails to fetch or if there are unexpected 
    Exceptions while calculating prices.
    """

    def __init__(self, exc: Exception, token: str, block: Optional[int], symbol: str):
        self.token = token
        """
        The token that caused the error.
        """

        self.block = block
        """
        The block that was queried when the error occurred.
        """

        self.exception = exc
        """
        The original :class:`~Exception` that was raised and wrapped with the yPriceMagicError.
        """

        detail = exc.__class__.__name__
        if not isinstance(self.exception, PriceError):
            detail += f'({exc})'
        # We do the import here to avoid a circular dependency
        from y import Network
        super().__init__(f"{detail} while fetching {Network.printable()} {symbol} {address} at block {block}")

class PriceError(Exception):
    """
    Raised when a queried price is not found.
    """

    def __init__(self, logger: logging.Logger, symbol: str):
        super().__init__(f"No price found for {symbol} {logger.address} at block {logger.block}")

class UnsupportedNetwork(ValueError):
    """
    Raised when an operation is attempted on an unsupported blockchain network.
    """

class NonStandardERC20(Exception):
    """
    Raised when an ERC20 token contract is expected but the provided address is not a standard ERC20 token.
    """

class CantFetchParam(Exception):
    pass


class TokenError(ValueError):
    """Raised when a token contract is not the correct contract type for the desired operation."""
    def __init__(self, token: AnyAddressType, desired_type: str, *optional_extra_args: Any):
        super().__init__(f"{token} is not a {desired_type}", *optional_extra_args)

# Explorer Exceptions

class _ExplorerError(Exception):
    """
    Base class for exceptions related to block explorer interactions.

    These errors are specific to issues encountered when interacting with blockchain explorers like Etherscan.
    """

class InvalidAPIKeyError(_ExplorerError):
    """
    Raised when the API key for the block explorer has been rejected.
    This typically occurs when making requests to a block explorer API with a missing, incorrect, or banned key.
    """
    _msg = "The block explorer for this network says your API key is invalid."
    def __init__(self, msg: str = ''):
        super().__init__(msg or self._msg)

# Contracts

class ContractNotVerified(_ExplorerError):
    """
    Raised when attempting to fetch the ABI for an unverified contract from a block explorer.
    """

class NoProxyImplementation(Exception):
    """
    Raised when the implementation address of a proxy contract cannot be determined.

    This may occur when trying to interact with proxy contracts that don't follow standard patterns.
    """

class MessedUpBrownieContract(Exception):
    """
    Raised when there's an issue initialized a Brownie contract instance,
    typically in the compilation step.
    """

    def __init__(self, address, *args: object) -> None:
        super().__init__(*args)
        # try to recache the contract
        try: BrownieContract.from_explorer(address)
        except CompilerError: pass # didn't work, oh well
        except Exception as e:
            if "invalid literal for int() with base 16: ''" in str(e): pass # didn't work, oh well
            elif "list index out of range" in str(e): pass # didn't work, oh well
            elif "pop from an empty deque" in str(e): pass # didn't work, oh well
            elif "'UsingForDirective' object has no attribute 'typeName'" in str(e): pass # didn't work, oh well
            elif contract_not_verified(e): pass # not verified, won't work
            else: raise


def contract_not_verified(e: Exception) -> bool:
    triggers = [
        'Contract source code not verified',
        'has not been verified',
    ]
    return any(trigger in str(e) for trigger in triggers)


# Pool detection

class NotAUniswapV2Pool(Exception):
    """
    Raised when a contract is incorrectly identified as a Uniswap V2 pool.
    """
    # NOTE: we use this exc to get the non-pool out of the pool singleton so it doesn't grow forever
    # TODO: Refactor this goofy thing out
    def __init__(self, non_pool: "UniswapV2Pool"):
        from y.prices.dex.uniswap.v2 import UniswapV2Pool
        UniswapV2Pool._ChecksumASyncSingletonMeta__instances[True].pop(non_pool.address, None)
        UniswapV2Pool._ChecksumASyncSingletonMeta__instances[False].pop(non_pool.address, None)
        super().__init__(non_pool.address)

class NotABalancerV2Pool(Exception):
    """
    Raised when a contract is incorrectly identified as a Balancer V2 pool.
    """

# Uni

class CantFindSwapPath(Exception):
    pass

class TokenNotFound(ValueError):
    """
    Raised when a specified token cannot be found in a given container.

    This is usually used when searching for a token in a liquidity pool.
    """
    
    def __init__(self, token, container):
        super().__init__(f"{token} is not in {container}")

# Calls

class CalldataPreparationError(Exception):
    """
    Raised when there's an error in preparing calldata for a contract interaction.
    """

class CallReverted(Exception):
    """
    Raised when a contract call is reverted.
    """


def call_reverted(e: Exception) -> bool:
    if isinstance(e, ContractLogicError):
        return True
    triggers = [
        'execution reverted',
        'No data was returned - the call likely reverted',
        'invalid opcode: opcode 0xfe not defined',
        'Tried to read 32 bytes.  Only got 0 bytes',
        'error processing call Revert',
        'invalid opcode: INVALID',
    ]
    return any(trigger in str(e) for trigger in triggers)


def continue_if_call_reverted(e: Exception) -> None:
    if call_reverted(e): return
    else: raise e


def out_of_gas(e: Exception) -> bool:
    return 'out of gas' in str(e) 

# Provider Exceptions:

class NodeNotSynced(Exception):
    pass



import threading
from collections import defaultdict
from typing import DefaultDict, Generic, TypeVar

from a_sync.a_sync._meta import ASyncMeta
from checksum_dict import ChecksumAddressDict
from checksum_dict.base import AnyAddressOrContract

T = TypeVar("T", bound=object)


class ChecksumASyncSingletonMeta(ASyncMeta, Generic[T]):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls.__instances: DefaultDict[bool, ChecksumAddressDict[T]] = defaultdict(ChecksumAddressDict)
        cls.__locks = defaultdict(lambda: defaultdict(threading.Lock))
        cls.__locks_lock: threading.Lock = threading.Lock()
    
    def __call__(cls, address: AnyAddressOrContract, *args, **kwargs) -> T:  # type: ignore
        # NOTE This will only work if you init your objects using a kwarg not a positional arg
        # TODO Make it work with posiional args
        address = str(address)
        is_sync = cls.__a_sync_instance_will_be_sync__(args, kwargs)
        try:
            instance = cls.__instances[is_sync][address]
        except KeyError:
            with cls.__get_address_lock(address, is_sync):
                # Try to get the instance again, in case it was added while waiting for the lock
                try:
                    instance =  cls.__instances[is_sync][address]
                except KeyError:
                    instance = super().__call__(address, *args, **kwargs)
                    cls.__instances[is_sync][address] = instance
            cls.__delete_address_lock(address, is_sync)
        assert instance.asynchronous is not is_sync, f"You must initialize your objects with 'asynchronous' specified as a kwarg, not a positional arg. {instance} {kwargs} {is_sync} {instance.asynchronous} {args} {kwargs}"
        return instance

    def __get_address_lock(cls, address: AnyAddressOrContract, is_sync: bool) -> threading.Lock:
        """ Makes sure the singleton is actually a singleton. """
        with cls.__locks_lock:
            return cls.__locks[is_sync][address]
    
    def __delete_address_lock(cls, address: AnyAddressOrContract, is_sync: bool) -> None:
        """ No need to maintain locks for initialized addresses. """
        with cls.__locks_lock:
            try:
                del cls.__locks[is_sync][address]
            except KeyError:
                pass



import logging
import threading
from typing import Dict, Optional

import a_sync
from brownie import chain
from cachetools import TTLCache, cached
from pony.orm import select
from y._db.decorators import a_sync_read_db_session, a_sync_write_db_session, log_result_count
from y._db.entities import Contract
from y._db.utils._ep import _get_get_token
from y._db.utils.utils import ensure_block
from y.datatypes import Address, Block


logger = logging.getLogger(__name__)

_get_get_token()

@a_sync_read_db_session
def get_deploy_block(address: str) -> Optional[int]:
    if deploy_block := known_deploy_blocks().pop(address, None):
        logger.debug('%s deploy block from cache: %s', address, deploy_block)
        return deploy_block
    get_token = _get_get_token()
    if deploy_block := get_token(address, sync=True).deploy_block:
        logger.debug('%s deploy block from cache: %s', address, deploy_block.number)
        return deploy_block.number
    logger.debug('%s deploy block not cached, fetching from chain', address)

async def _set_deploy_block(address: str, deploy_block: int) -> None:
    await ensure_block(deploy_block)
    return await __set_deploy_block(address, deploy_block)

set_deploy_block = a_sync.ProcessingQueue(_set_deploy_block, num_workers=10, return_data=False)

@a_sync_write_db_session
def __set_deploy_block(address: str, deploy_block: int) -> None:
    from y._db.utils._ep import _get_get_token
    get_token = _get_get_token()
    get_token(address, sync=True).deploy_block = (chain.id, deploy_block)
    logger.debug('deploy block cached for %s: %s', address, deploy_block)


# startup caches

@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@log_result_count("deploy blocks")
def known_deploy_blocks() -> Dict[Address, Block]:
    """cache and return all known contract deploy blocks for this chain to minimize db reads"""
    return dict(select((c.address, c.deploy_block.number) for c in Contract if c.chain.id == chain.id and c.deploy_block.number))




import logging
import threading
from contextlib import suppress
from decimal import Decimal, InvalidOperation
from typing import Dict, Optional

import a_sync
from brownie import chain
from cachetools import TTLCache, cached
from pony.orm import select

from y import constants
from y._db.decorators import a_sync_read_db_session, log_result_count, retry_locked
from y._db.entities import Price, insert
from y._db.utils.token import ensure_token
from y._db.utils.utils import ensure_block
from y.datatypes import Address


logger = logging.getLogger(__name__)

@a_sync_read_db_session
def get_price(address: str, block: int) -> Optional[Decimal]:
    if address == constants.EEE_ADDRESS:
        address = constants.WRAPPED_GAS_COIN
    if price := known_prices_at_block(block).pop(address, None):
        logger.debug("found %s block %s price %s in ydb", address, block, price)
        return price
    if (price := Price.get(token = (chain.id, address), block = (chain.id, block))) and (price:=price.price):
        logger.debug("found %s block %s price %s in ydb", address, block, price)
        return price

@retry_locked
async def _set_price(address: str, block: int, price: Decimal) -> None:
    await ensure_block(block, sync=False)
    if address == constants.EEE_ADDRESS:
        address = constants.WRAPPED_GAS_COIN
    await ensure_token(address, sync=False)
    with suppress(InvalidOperation): # happens with really big numbers sometimes. nbd, we can just skip the cache in this case.
        await insert(
            type = Price,
            block = (chain.id, block),
            token = (chain.id, address),
            price = Decimal(price),
            sync = False,
        )
        logger.debug("inserted %s block %s price to ydb: %s", address, block, price)

set_price = a_sync.ProcessingQueue(_set_price, num_workers=50, return_data=False)

@cached(TTLCache(maxsize=1_000, ttl=5*60), lock=threading.Lock())
@log_result_count("prices", ["block"])
def known_prices_at_block(number: int) -> Dict[Address, Decimal]:
    """cache and return all known prices at block `number` to minimize db reads"""
    return dict(select((p.token.address, p.price) for p in Price if p.block.chain.id == chain.id and p.block.number == number))



import asyncio
from decimal import Decimal
from typing import Tuple

from a_sync import a_sync
from async_lru import alru_cache

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20
from y.contracts import Contract, has_method
from y.datatypes import Address, Block
from y.exceptions import ContractNotVerified


try:
    PENDLE_ORACLE = Contract("0x9a9Fa8338dd5E5B2188006f1Cd2Ef26d921650C2")
except ContractNotVerified:
    PENDLE_ORACLE = None
    
TWAP_DURATION = 1


@a_sync("sync")
async def is_pendle_lp(token: Address) -> bool:
    """
    Determines if the given token address is a Pendle LP token.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is a Pendle LP token, False otherwise.

    Example:
        >>> is_pendle = is_pendle_lp("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23")
        >>> print(is_pendle)
        True
    """
    return PENDLE_ORACLE is not None and await has_method(token, 'readTokens()(address,address,address)', sync=False)

@alru_cache(maxsize=None)
async def get_tokens(lp_token: Address) -> Tuple[str, str, str]:
    """
    Retrieves the addresses of the tokens in a Pendle LP token.

    This function is cached to improve performance for repeated calls.

    Args:
        lp_token: The address of the Pendle LP token.

    Returns:
        A tuple containing the addresses of the SY token, PT token, and YT token.

    Example:
        >>> tokens = await get_tokens("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23")
        >>> print(tokens)
        ('0x...', '0x...', '0x...')  # Addresses of SY, PT, and YT tokens
    """
    lp_token = await Contract.coroutine(lp_token)
    return await lp_token.readTokens

@a_sync("sync")
async def get_lp_price(token: Address, block: Block = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Decimal:
    """
    Calculates the price of a Pendle LP token.

    Args:
        token: The address of the Pendle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the LP token in USD.

    Example:
        >>> price = get_lp_price("0x1b92b5242301ce4a8c73cc3ef0d6dee33a3a5b23", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Pendle LP token in USD

    Note:
        This function uses the Pendle Oracle contract to get the LP to asset rate and then
        calculates the USD price based on the underlying asset's price.
    """
    tokens = await get_tokens(token)
    # NOTE: we might not need this, leave it commented out for now
    #names = await asyncio.gather(*[ERC20(t, asynchronous=True).name for t in tokens])
    #if any("DAI" in name for name in names):
    #    use_asset = True
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    #elif any("crvUSD" in name for name in (ERC20(t) for t in tokens)):
    #    use_asset = True
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    #else:
    #    use_asset = False
    #    rate = await PENDLE_ORACLE.getLpToAssetRate.coroutine(token, twap_duration, block_identifier=block)
    sy_token, p_token, y_token = tokens
    sy, rate = await asyncio.gather(
        Contract.coroutine(sy_token),
        PENDLE_ORACLE.getLpToAssetRate.coroutine(token, TWAP_DURATION, block_identifier=block),
    )
    _, asset, decimals = await sy.assetInfo
    rate /= Decimal(10**decimals)
    return rate * await ERC20(asset, asynchronous=True).price(block=block, skip_cache=skip_cache)




import asyncio
import logging
from decimal import Decimal
from typing import Optional

import a_sync

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic
from y.utils import gather_methods

logger = logging.getLogger(__name__)

@a_sync.a_sync(default='sync', ram_cache_ttl=5*60)
async def is_ib_token(token: AnyAddressType) -> bool:
    """
    Determines if the given token address is an Iron Bank token.

    Args:
        token: The address of the token to check.

    Returns:
        True if the token is an Iron Bank token, False otherwise.

    Example:
        >>> is_ib = is_ib_token("0x41c84c0e2EE0b740Cf0d31F63f3B6F627DC6b393")
        >>> print(is_ib)
        True
    """
    return await has_methods(token, ('debtShareToVal(uint)(uint)','debtValToShare(uint)(uint)','token()(address)','totalToken()(uint)','totalSupply()(uint)'), sync=False)

@a_sync.a_sync(default='sync')
async def get_price(token: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> UsdPrice:
    """
    Calculates the price of an Iron Bank token.

    Args:
        token: The address of the Iron Bank token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obbj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the Iron Bank token in USD.

    Example:
        >>> price = get_price("0x41c84c0e2EE0b740Cf0d31F63f3B6F627DC6b393", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Iron Bank token in USD

    Note:
        This function calculates the price by determining the share price of the Iron Bank token
        relative to its underlying asset, and then multiplying by the price of the underlying asset.
    """
    address = convert.to_address(token)
    token, total_bal, total_supply = await gather_methods(address, ['token', 'totalToken', 'totalSupply'], block=block)
    token_scale, pool_scale = await asyncio.gather(ERC20(token, asynchronous=True).scale, ERC20(address, asynchronous=True).scale)
    total_bal /= Decimal(token_scale)
    total_supply /= Decimal(pool_scale)
    share_price = total_bal / total_supply
    token_price = await magic.get_price(token, block, skip_cache=skip_cache, sync=False)
    return share_price * token_price




import logging
import threading
from functools import lru_cache
from typing import Dict, Optional, Set

import a_sync
from brownie import chain, convert
from cachetools import TTLCache, cached
from pony.orm import commit, db_session, select

from y import constants
from y._db.decorators import a_sync_read_db_session, a_sync_write_db_session, log_result_count
from y._db.entities import Address, Token, insert
from y._db.exceptions import EEEError
from y._db.utils._ep import _get_get_token
from y.erc20 import decimals

logger = logging.getLogger(__name__)

@a_sync_read_db_session
def get_token(address: str) -> Token:
    address = convert.to_address(address)
    if address == constants.EEE_ADDRESS:
        raise EEEError(f"cannot create token entity for {constants.EEE_ADDRESS}")
    while True:
        if entity := Address.get(chain=chain.id, address=address):
            if isinstance(entity, Token):
                return entity
            entity.delete()
            commit()
        return insert(type=Token, chain=chain.id, address=address) or Token.get(chain=chain.id, address=address)

@a_sync.a_sync(default='sync', ram_cache_maxsize=None)
def ensure_token(address: str) -> None:
    return _ensure_token(address)

@lru_cache(maxsize=None)
@db_session
def _ensure_token(address: str) -> None:
    """We can't wrap a `_Wrapped` object with `a_sync` so we have this helper fn"""
    if address not in known_tokens():
        get_token = _get_get_token()
        get_token(address, sync=True)

@a_sync_read_db_session
def get_bucket(address: str) -> Optional[str]:
    if address == constants.EEE_ADDRESS:
        return
    if (bucket := known_buckets().pop(address, None)) is None:
        get_token = _get_get_token()
        bucket = get_token(address, sync=True).bucket
    if bucket:
        logger.debug("found %s bucket %s in ydb", address, bucket)
    return bucket

@a_sync_write_db_session
def _set_bucket(address: str, bucket: str) -> None:
    if address == constants.EEE_ADDRESS:
        return
    get_token = _get_get_token()
    get_token(address, sync=True).bucket = bucket
    logger.debug("updated %s bucket in ydb: %s", address, bucket)

set_bucket = a_sync.ProcessingQueue(_set_bucket, num_workers=10, return_data=False)

@a_sync_read_db_session
def get_symbol(address: str) -> Optional[str]:
    if (symbol := known_symbols().pop(address, None)) is None:
        get_token = _get_get_token()
        symbol = get_token(address, sync=True).symbol
    if symbol:
        logger.debug("found %s symbol %s in ydb", address, symbol)
    return symbol

def set_symbol(address: str, symbol: str):
    if not symbol:
        raise ValueError("`symbol` is required")
    if not isinstance(symbol, str):
        raise TypeError(f"`symbol` must be a string. You passed {symbol}")
    a_sync.create_task(
        coro=_set_symbol(address, symbol), 
        name=f"set_symbol {symbol} for {address}",
        skip_gc_until_done=True,
    )

@a_sync_read_db_session
def get_name(address: str) -> Optional[str]:
    if (name := known_names().pop(address, None)) is None:
        get_token = _get_get_token()
        name = get_token(address, sync=True).name
    if name:
        logger.debug("found %s name %s in ydb", address, name)
    return name

def set_name(address: str, name: str) -> None:
    if not name:
        raise ValueError("`name` is required")
    if not isinstance(name, str):
        raise TypeError(f"`name` must be a string. You passed {name}")
    a_sync.create_task(
        coro=_set_name(address, name), 
        name=f"set_name {name} for {address}",
        skip_gc_until_done=True,
    )

async def get_decimals(address: str) -> int:
    d = await _get_token_decimals(address)
    if d is None:
        d = await decimals(address, sync=False)
        if d:
            if d > 2147483647:
                # raise this here so the task doesnt break
                # TODO debug where this comes from
                raise ValueError(f"Value {d} of attr Token.decimals is greater than the maximum allowed value 2147483647")
            a_sync.create_task(coro=set_decimals(address, d), name=f"set_decimals {address}", skip_gc_until_done=True)
    return d

@a_sync_write_db_session
def set_decimals(address: str, decimals: int) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).decimals = decimals
    logger.debug("updated %s decimals in ydb: %s", address, decimals)

@a_sync_write_db_session
def _set_symbol(address: str, symbol: str) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).symbol = symbol
    logger.debug("updated %s symbol in ydb: %s", address, symbol)
    
@a_sync_write_db_session
def _set_name(address: str, name: str) -> None:
    get_token = _get_get_token()
    get_token(address, sync=True).name = name
    logger.debug("updated %s name in ydb: %s", address, name)

@a_sync_read_db_session
def _get_token_decimals(address: str) -> Optional[int]:
    if (decimals := known_decimals().pop(address, None)) is None:
        get_token = _get_get_token()
        decimals = get_token(address, sync=True).decimals
    if decimals:
        logger.debug("found %s decimals %s in ydb", address, decimals)
        return decimals


# startup caches
    
@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@db_session
@log_result_count("tokens")
def known_tokens() -> Set[str]:
    """cache and return all known Tokens for this chain to minimize db reads"""
    return set(select(t.address for t in Token if t.chain.id == chain.id))

@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@log_result_count("buckets")
def known_buckets() -> Dict[str, str]:
    """cache and return all known token buckets for this chain to minimize db reads"""
    return dict(select((t.address, t.bucket) for t in Token if t.chain.id == chain.id and t.bucket))

@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@log_result_count("token decimals")
def known_decimals() -> Dict[Address, int]:
    """cache and return all known token decimals for this chain to minimize db reads"""
    return dict(select((t.address, t.decimals) for t in Token if t.chain.id == chain.id and t.decimals))

@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@log_result_count("token symbols")
def known_symbols() -> Dict[Address, str]:
    """cache and return all known token symbols for this chain to minimize db reads"""
    return dict(select((t.address, t.symbol) for t in Token if t.chain.id == chain.id and t.symbol))

@cached(TTLCache(maxsize=1, ttl=60*60), lock=threading.Lock())
@log_result_count("token names")
def known_names() -> Dict[Address, str]:
    """cache and return all known token names for this chain to minimize db reads"""
    return dict(select((t.address, t.name) for t in Token if t.chain.id == chain.id and t.name))


import asyncio
import logging
from contextlib import suppress
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.convert.datatypes import EthAddress
from brownie.exceptions import VirtualMachineError
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.constants import dai, usdc, wbtc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (Address, AddressOrContract, AnyAddressType, Block,
                         Pool, UsdPrice, UsdValue)
from y.networks import Network
from y.prices import magic
from y.prices.dex.balancer._abc import BalancerABC, BalancerPool
from y.utils.cache import optional_async_diskcache

EXCHANGE_PROXY = {
    Network.Mainnet: '0x3E66B66Fd1d0b02fDa6C811Da9E0547970DB2f21',
}.get(chain.id)

SCALES_TO_TRY = [1.0, 0.5, 0.1]
TOKENOUTS_TO_TRY = [weth, dai, usdc, wbtc]

logger = logging.getLogger(__name__)

async def _calc_out_value(token_out: AddressOrContract, total_outout: int, scale: float, block: int, skip_cache: bool = ENVS.SKIP_CACHE) -> float:
    out_scale, out_price = await asyncio.gather(ERC20(token_out, asynchronous=True).scale, magic.get_price(token_out, block, skip_cache=skip_cache, sync=False))
    return (total_outout / out_scale) * float(out_price) / scale

class BalancerV1Pool(BalancerPool):
    @a_sync.aka.cached_property
    @stuck_coro_debugger
    @optional_async_diskcache
    async def tokens(self) -> List[ERC20]:
        contract = await Contract.coroutine(self.address)
        return [ERC20(token, asynchronous=self.asynchronous) for token in await contract.getFinalTokens]
    __tokens__: HiddenMethodDescriptor[Self, List[ERC20]]

    @stuck_coro_debugger
    async def get_tvl(self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdValue]:
        token_balances = await self.get_balances(block=block, sync=False)
        good_balances = {
            token: balance
            for token, balance
            in token_balances.items()
            if await token.price(block=block, return_None_on_failure=True, skip_cache=skip_cache, sync=False) is not None
        }
        
        prices = await ERC20.price.map(good_balances, block=block, return_None_on_failure=True, skip_cache=skip_cache).values()

        # in case we couldn't get prices for all tokens, we can extrapolate from the prices we did get
        good_value = sum(balance * Decimal(price) for balance, price in zip(good_balances.values(),prices))
        if len(good_balances):
            return good_value / len(good_balances) * len(token_balances)
        return None
    
    @stuck_coro_debugger
    async def get_balances(self, block: Optional[Block] = None) -> Dict[ERC20, Decimal]:
        return await a_sync.map(self.get_balance, self.__tokens__, block=block or 'latest')

    @stuck_coro_debugger
    async def get_balance(self, token: AnyAddressType, block: Block) -> Decimal:
        balance, scale = await asyncio.gather(
            self.check_liquidity(str(token), block, sync=False),
            ERC20(token, asynchronous=True).scale,
        )
        return Decimal(balance) / scale

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=10_000, ram_cache_ttl=10*60)
    async def check_liquidity(self, token: Address, block: Block) -> int:
        if block < await self.deploy_block(sync=False):
            return 0
        contract = await Contract.coroutine(self.address)
        return await contract.getBalance.coroutine(token, block_identifier=block)
    

class BalancerV1(BalancerABC[BalancerV1Pool]):

    _pool_type = BalancerV1Pool

    _check_methods = ("getCurrentTokens()(address[])", "getTotalDenormalizedWeight()(uint)", "totalSupply()(uint)")

    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        self.exchange_proxy = Contract(EXCHANGE_PROXY) if EXCHANGE_PROXY else None
    
    @stuck_coro_debugger
    async def get_token_price(self, token_address: AddressOrContract, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        if block is not None and block < await contract_creation_block_async(self.exchange_proxy, True):
            return None
        for scale in SCALES_TO_TRY:
            # Can we get an output if we try smaller size? try consecutively smaller
            if output := await self.get_some_output(token_address, block=block, scale=scale, sync=False):
                return await _calc_out_value(*output, scale, block=block, skip_cache=skip_cache)
    
    @stuck_coro_debugger
    async def check_liquidity_against(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Optional[int]:

        amount_in = await ERC20(token_in, asynchronous=True).scale * scale
        with suppress(ValueError, VirtualMachineError, ContractLogicError):
            # across various dep versions we get these various excs
            view_split_exact_in = await self.exchange_proxy.viewSplitExactIn.coroutine(
                token_in,
                token_out,
                amount_in,
                32, # NOTE: 32 is max
                block_identifier = block,
            )
            return view_split_exact_in['totalOutput']
    
    @stuck_coro_debugger
    async def get_some_output(
        self,
        token_in: AddressOrContract,
        scale: int = 1,
        block: Optional[Block] = None
        ) -> Optional[Tuple[EthAddress,int]]:
        for token_out in TOKENOUTS_TO_TRY:
            if output := await self.check_liquidity_against(token_in, token_out, block=block, scale=scale, sync=False):
                return token_out, output

    @stuck_coro_debugger
    async def check_liquidity(self, token: Address, block: Block, ignore_pools: Tuple[Pool, ...] = ()) -> int:
        pools = []
        pools = [pool for pool in pools if pool not in ignore_pools]
        return await BalancerV1Pool.check_liquidity.max(pools, token=token, block=block, sync=False) if pools else 0




from random import randint

from brownie.convert.datatypes import EthAddress
from y import convert
from y.classes.common import ContractBase
from y.constants import WRAPPED_GAS_COIN
from y.contracts import Contract


def test_contract_singleton():
    token = WRAPPED_GAS_COIN
    variations = [token.lower(), token.upper(), convert.to_address(token), Contract(token), EthAddress(token.lower())]
    for variation in variations:
        assert ContractBase(WRAPPED_GAS_COIN) == ContractBase(variation), 'ContractSingleton metaclass is not working.'




import logging
from typing import List, Optional, Union

import a_sync
from a_sync.a_sync.property import HiddenMethodDescriptor
from brownie import chain
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y import exceptions
from y._decorators import stuck_coro_debugger
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices.dex.balancer._abc import BalancerABC
from y.prices.dex.balancer.v1 import BalancerV1
from y.prices.dex.balancer.v2 import BalancerV2
from y.utils.cache import optional_async_diskcache


logger = logging.getLogger(__name__)

class BalancerMultiplexer(a_sync.ASyncGenericBase):
    def __init__(self, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
    
    @a_sync.aka.property
    async def versions(self) -> List[Union[BalancerV1, BalancerV2]]:
        return [v async for v in a_sync.as_completed([self.__v1__, self.__v2__], aiter=True) if v]
    __versions__: HiddenMethodDescriptor[Self, List[Union[BalancerV1, BalancerV2]]]

    @a_sync.aka.cached_property
    async def v1(self) -> Optional[BalancerV1]:
        try:
            return BalancerV1(asynchronous=self.asynchronous)
        except ImportError:
            return None
    __v1__: HiddenMethodDescriptor[Self, Optional[BalancerV1]]
    
    @a_sync.aka.cached_property
    async def v2(self) -> Optional[BalancerV2]:
        try:
            return BalancerV2(asynchronous=self.asynchronous)
        except ImportError: 
            return None
    __v2__: HiddenMethodDescriptor[Self, Optional[BalancerV2]]

    @stuck_coro_debugger
    @optional_async_diskcache
    async def is_balancer_pool(self, token_address: AnyAddressType) -> bool:
        try:
            await self.get_version(token_address)
            return True
        except exceptions.TokenError:
            return False
    
    @stuck_coro_debugger
    async def get_pool_price(self, token_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        balancer: BalancerABC = await self.get_version(token_address)
        logger.debug("pool %s is from %s", token_address, balancer)
        return UsdPrice(await balancer.get_pool_price(token_address, block, skip_cache=skip_cache, sync=False))
            
    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_price(self, token_address: AnyAddressType, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE) -> Optional[UsdPrice]:
        if await self.is_balancer_pool(token_address, sync=False):
            return await self.get_pool_price(token_address, block=block, skip_cache=skip_cache, sync=False)

        price = None
        
        if ( # NOTE: Only query v2 if block queried > v2 deploy block plus some extra blocks to build up liquidity
            (chain.id == Network.Mainnet and (not block or block > 12272146 + 100000))
            or (chain.id == Network.Fantom and (not block or block > 16896080))
        ):  # TODO: refactor this out
            v2 = await self.__v2__
            if price := await v2.get_token_price(token_address, block, skip_cache=skip_cache, sync=False):
                logger.debug("balancer v2 -> $%s", price)
                return price

        if not price and chain.id == Network.Mainnet:   
            v1 = await self.__v1__
            if price := await v1.get_token_price(token_address, block, skip_cache=skip_cache, sync=False):
                logger.debug("balancer v1 -> $%s", price)
                return price
    
    # cached forever because not many items
    @a_sync.a_sync(cache_type="memory", ram_cache_ttl=None)  
    async def get_version(self, token_address: AnyAddressType) -> BalancerABC:
        for v in await self.__versions__:
            if await v.is_pool(token_address, sync=False):
                return v
        raise exceptions.TokenError(token_address, "Balancer pool")

balancer_multiplexer = BalancerMultiplexer(asynchronous=True)



import os
from contextlib import suppress

from brownie import network
from brownie.network.contract import _explorer_tokens

if "base" not in _explorer_tokens:
    _explorer_tokens["base"] = "BASESCAN_TOKEN"
if "optimistic" not in _explorer_tokens:
    _explorer_tokens["optimistic"] = "OPTIMISMSCAN_TOKEN"

class NetworkNotSpecified(Exception):
    pass
    
if not network.is_connected():
    if 'BROWNIE_NETWORK_ID' not in os.environ:
        raise NetworkNotSpecified('In order to use pricemagic outside of a brownie project directory, you will need to set $BROWNIE_NETWORK_ID environment variable with the id of your preferred brownie network connection.')
    network.connect(os.environ['BROWNIE_NETWORK_ID'])

# for backwards-compatability
from dank_mids import dank_web3 as dank_w3

from y import time
from y.classes.common import ERC20
from y.constants import EEE_ADDRESS, WRAPPED_GAS_COIN, dai, usdc, wbtc, weth
from y.contracts import (Contract, Contract_erc20, Contract_with_erc20_fallback, 
                         contract_creation_block, contract_creation_block_async, 
                         has_method, has_methods)
from y.datatypes import Address, Block
from y.erc20 import decimals, totalSupply, totalSupplyReadable
from y.exceptions import (CalldataPreparationError, CallReverted,
                          ContractNotVerified, MessedUpBrownieContract,
                          NonStandardERC20, NotABalancerV2Pool,
                          NotAUniswapV2Pool, PriceError, UnsupportedNetwork)
from y import exceptions
from y.networks import Network
from y.prices import magic
from y.prices.magic import get_price, get_prices, map_prices
from y.prices.utils import check_bucket
from y.time import get_block_at_timestamp, get_block_timestamp, get_block_timestamp_async
from y.utils.logging import enable_debug_logging
from y.utils.middleware import setup_getcode_cache_middleware
from y.utils.multicall import fetch_multicall
from y.utils.raw_calls import balanceOf, raw_call

setup_getcode_cache_middleware()

__all__ = [
    ### you can reach the below functions, classes, and variables using ###
    ###  `y.__name__` or `from y import __name__`.  ###

    # prices
    'get_price',
    'get_prices',
    'map_prices',
    'check_bucket',

    # erc20
    'ERC20',

    # contract stuff
    'Contract',
    'contract_creation_block',
    'contract_creation_block_async',
    'has_method',
    'has_methods',

    # network stuff
    'Network',

    # constants
    'EEE_ADDRESS',
    'WRAPPED_GAS_COIN',
    'weth',
    'dai',
    'usdc',
    'wbtc',

    # magic
    'magic',

    # raw calls
    'raw_call',

    # time
    'time',
    'get_block_at_timestamp', 
    'get_block_timestamp', 
    'get_block_timestamp_async',
]

with suppress(ModuleNotFoundError):
    """If eth_portfolio is also installed in this env, we will use its extended version of our db schema"""
    from eth_portfolio._db import entities as _db_entities
